(*  Title:      Typed_ASP.thy
    ID:         $Id: Typed_ASP.thy,v 1.9 2009/01/09 17:43:24 flokam Exp $
    Author:     Ludovic Henrio, Florian Kammuller and Henry Sudhof
                2006

    Note: 
       From: Typed_Sigma.thy (Florian Kammuller and Henry Sudhof)
       plus:
        - type for references to futures and active objects
        - typing rule for activities
        - proof of subject reduction adapted to ASP !!!
*)

header {* First Order Types for Sigma terms with aspect labels *}

theory Typed_ASP imports CyclesOfFutures begin

subsection {*Type Safety*}

subsection {* Environments *}

constdefs
   shift :: "('a list) \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a list"    ("_<_:_>" [90, 0, 0] 91)
   "l<i:a> == list_insert l i a"



(* \<lambda>j. if j < i then e j else if j = i then a else e (j - 1)"


               (if (i = 0) then (a#l#ls) else 
               (if ( (length (l#ls)) = i) then ((l#ls)@[a]) 
                  else ((take i (l#ls))@(a#(drop i (l#ls))   ))) 
*)
syntax (xsymbols)
   shift :: "('a list) \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a list"    ("_\<langle>_:_\<rangle>" [90, 0, 0] 91)
syntax (HTML output)
   shift :: "('a list) \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a list"   ("_\<langle>_:_\<rangle>" [90, 0, 0] 91)


lemma shift_eq [simp]: "i = j \<and>  j \<le> length e \<Longrightarrow> (e\<langle>i:T\<rangle>)!(j) = T"
apply(unfold shift_def)
apply(auto)
done

lemma shift_gt [simp]: "\<lbrakk>j < i; i \<le> length e\<rbrakk> \<Longrightarrow> (e\<langle>i:T\<rangle>)! j = e!j"
apply(unfold shift_def)
apply(auto)
done


lemma shift_lt [simp]: "\<lbrakk>i \<le>  j\<rbrakk> \<Longrightarrow> (e\<langle>i:T\<rangle>)! (Suc j) = e!(j)"
apply(unfold shift_def)
apply(auto)
done

lemma shift_lt1 [simp]: "\<lbrakk>i <  j; i \<le> length e\<rbrakk> \<Longrightarrow> (e\<langle>i:T\<rangle>)!(j) = e!(j - 1)"
apply(unfold shift_def)
apply(auto)

apply(subgoal_tac "Suc (j - 1) = j")
apply(simp)
apply(induct j)
apply(auto)
done


lemma shift_lt2: "\<lbrakk>i <   j\<rbrakk> \<Longrightarrow> (e\<langle>i:T\<rangle>)!Suc(j - 1) = e!(j - 1)"
apply(subst  shift_lt)
apply(auto)
done


lemma shift_lt3 [simp]: "\<lbrakk>i <  j\<rbrakk> \<Longrightarrow> (e\<langle>i:T\<rangle>)!(j) = e!(j - 1)"
apply(auto)
apply(induct j)
apply(auto)
done



lemma shift_commute [simp]: "\<lbrakk>i \<le> length e\<rbrakk> \<Longrightarrow> e\<langle>i:U\<rangle>\<langle>0:T\<rangle> = e\<langle>0:T\<rangle>\<langle>Suc i:U\<rangle>"
apply(unfold shift_def)
apply(auto)
done



lemma shift_length [simp] : " \<forall> x  .length ( e<x:b> ) = Suc (length e)"
apply(auto)

apply(unfold shift_def)
apply(induct e)
apply(auto)
apply(case_tac x)
apply(auto)
done

subsection {* Types and typing rules *}

(*configuration type *)
datatype Ctype = TConfig "ActivityRef ~=> type" "FutureRef~=> type"

consts
  type_get :: "type \<Rightarrow> Label \<Rightarrow> (type\<times>type)"  ("_!_")
primrec
 " (Object l)!n = the (l n) "


consts
  type_get_opt :: "type \<Rightarrow> Label \<Rightarrow>(type\<times>type)  option"  ("_!!_")
primrec
 " (Object l)!!n =  (l n) "

consts
  do :: "type \<Rightarrow> (Label set)"  
primrec
 "do (Object l) = (dom l)"

inductive
  typing :: "[Ctype, ((type\<times>type) list), dB, type] \<Rightarrow> bool" ("_ , _ \<turnstile> _ : _" [50, 50, 50] 50)
where
 T_Self [intro!]:  "\<lbrakk>x < length env; (env!x) = (T,T') \<rbrakk> \<Longrightarrow> Tconf, env \<turnstile> Var (Self x) : T" |

 T_Param [intro!]: "\<lbrakk>x < length env; (env!x) = (T,T') \<rbrakk> \<Longrightarrow> Tconf, env \<turnstile> Var (Param x) : T'" |

 T_Obj  [ intro!]: "\<lbrakk> dom b = do B; 
                      \<forall>l \<in>  (do B) . Tconf,env<0:(B,(fst(B!l)))>  \<turnstile> (the (b l)) : (snd(B!l)) \<rbrakk> 
                    \<Longrightarrow> Tconf, env \<turnstile> (Obj b B) :  B" |

 T_Call [ intro!]: "\<lbrakk> Tconf, env \<turnstile> a : A; l \<in> do A; A!l=(B,T);Tconf, env \<turnstile> b : B \<rbrakk> 
                            \<Longrightarrow> Tconf, env \<turnstile> (Call a l b) : T" |
 T_Upd  [ intro!]: "\<lbrakk> Tconf, env \<turnstile> a : A; l \<in> (do A) ; 
                         Tconf,env<0:(A,(fst(A!l)))>  \<turnstile> n : (snd(A!l)) \<rbrakk> 
                            \<Longrightarrow> Tconf, env \<turnstile>  (Upd a l n) : A" |
 T_Active [intro!]:"\<lbrakk>Tconf, env \<turnstile> a : A \<rbrakk> \<Longrightarrow> Tconf,env \<turnstile>  (Active a) : A" |
 T_ActRef [intro!]:"\<lbrakk> amap B = Some T;B\<in>dom amap\<rbrakk> 
                        \<Longrightarrow> TConfig amap futmap,env \<turnstile>  (ActRef B) : T" |
 T_FutRef [intro!]:"\<lbrakk> futmap fi = Some T;fi \<in> dom futmap\<rbrakk> 
                        \<Longrightarrow> TConfig amap futmap,env \<turnstile>  (FutRef fi) : T" 
 
inductive_cases typing_elims [elim!]:
  "c, e \<turnstile> Obj b B: T"
  "c, e \<turnstile> Var (Self i) : T"
  "c, e \<turnstile> Var (Param i) : T"
  "c, e \<turnstile> Call a l b : T"
  "c, e \<turnstile> Upd a l n  : T"
  "c, e \<turnstile> Active a : T"
  "c, e \<turnstile> ActRef B : T"
  "c, e \<turnstile> ActFut fi : T";


(*typing configuration, short but intense: 
  - verify bijection between domains 
  - typecheck each AO and future
*)
inductive  Ctyping :: "[configuration, Ctype] \<Rightarrow> bool" ("\<turnstile>_ : _" [50] 50)
where
  T_conf: "\<lbrakk>dom amap  = dom c; 
            dom futmap = Union{dom m|m. \<exists>A a. c A=Some (m,a)};
             \<forall>A \<in> dom c. (TConfig amap futmap,[]\<turnstile>c.\<^sub>AA:the (amap A))\<and>
                       (\<forall>fi\<in>dom (c.\<^sub>RA). (TConfig amap futmap),[]\<turnstile>(c.\<^sub>FA<fi>):the(futmap fi))\<rbrakk>
  \<Longrightarrow>   \<turnstile> c : TConfig amap futmap"

primrec typings :: "Ctype \<Rightarrow> ((type\<times>type) list) \<Rightarrow> dB  list \<Rightarrow> type list \<Rightarrow> bool" ("_ , _ ||- _ : _" [50, 50, 50] 50)
where
 "(c, e ||- [] : Ts) = (Ts = [])" |
 "(c, e ||- t # ts : Ts) = 
   (case Ts of
     [] \<Rightarrow> False
    | T # Ts \<Rightarrow> c, e \<turnstile> t : T \<and> c, e ||- ts : Ts)"

subsection {* Lists of types *}

lemma lists_typings:
    "\<And>Ts. c,e ||- ts : Ts \<Longrightarrow> ts \<in> lists {t. \<exists>T. (c, e \<turnstile> t : T)}"
  apply (induct ts)
  apply(simp)
  apply force
   apply (case_tac Ts)
     apply simp
(* proof below had to be added because of the additional parameter c
 was force before :-( *)
apply simp;
apply (erule conjE);
apply (rule lists.intros(2))
apply simp+;
apply (rule_tac x=aa in exI);
apply simp+;
done

lemma types_snoc: "\<And>Ts. c,e ||- ts : Ts \<Longrightarrow> c,e \<turnstile> t : T \<Longrightarrow> c,e ||- ts @ [t] : Ts @ [T]"
  apply (induct ts)
  apply simp
  apply (case_tac Ts)
  apply simp+
  done

lemma types_snoc_eq: "\<And>Ts. c,e ||- ts @ [t] : Ts @ [T] =
  (c,e ||- ts : Ts \<and> c,e \<turnstile> t : T)"
  apply (induct ts)
  apply (simp)+
  apply (case_tac Ts)
  apply simp+
  apply (case_tac Ts)
  apply (case_tac "ts @ [t]")
  apply simp+
  done

lemma rev_exhaust2 [case_names Nil snoc, extraction_expand]:
  "(xs = [] \<Longrightarrow> P) \<Longrightarrow> (\<And>ys y. xs = ys @ [y] \<Longrightarrow> P) \<Longrightarrow> P"
  -- {* Cannot use @{text rev_exhaust} from the @{text List}
    theory, since it is not constructive *}
  apply (subgoal_tac "\<forall>ys. xs = rev ys \<longrightarrow> P")
  apply (erule_tac x="rev xs" in allE)
  apply simp
  apply (rule allI)
  apply (rule impI)
  apply (case_tac ys)
  apply simp
  apply simp
  apply atomize
  apply (erule allE)+
  apply (erule mp, rule conjI)
  apply (rule refl)+
  done

lemma types_snocE: "c,e ||- ts @ [t] : Ts \<Longrightarrow>
  (\<And>Us U. Ts = Us @ [U] \<Longrightarrow> c,e ||- ts : Us \<Longrightarrow> c,e \<turnstile> t : U \<Longrightarrow> P) \<Longrightarrow> P"
  apply (cases Ts rule: rev_exhaust2)
  apply simp
  apply (case_tac "ts @ [t]")
  apply (simp add: types_snoc_eq)+
  apply iprover
  done


subsection {* Lifting preserves well-typedness *}

lemma lift_option_dom [simp] : " \<forall> b. dom (\<lambda>l. lift_option (i) (b l)) = dom b"
apply(auto)
apply(case_tac "x \<in>  dom b")
apply(auto)
done


lemma lift_lift_option [rule_format] : "\<forall> l \<in> dom b .  (lift (the (b l)) x) = the (lift_option x ( b l))"
by auto;

lemma lift_type [intro!]: "  c,e \<turnstile> t : T  \<Longrightarrow> (\<And> U V. \<forall> i \<le> length e .  (c,e\<langle>i:(U,V)\<rangle> \<turnstile> lift t i : T))"
  apply(induct set: typing)
  apply(force)
  apply(force)
  apply(rule allI)
  apply(rule impI)
  apply(simp)
  apply(simp add : lift_lift_option)
  apply(rule typing.intros)
  apply(force)
  apply(rule ballI)
  apply(erule_tac x=l in ballE)
  apply(erule conjE)

  apply(erule_tac x=U in allE) 
  apply(erule_tac x=V in allE) 
  apply(erule_tac x="Suc i" in allE) 
  apply(erule impE)
  apply(arith)
  apply(simp add : lift_lift_option)
  apply(simp)
  apply(simp)
  apply(rule allI)
  apply(rule impI)
  apply(drule_tac x=U in meta_spec) 
  apply(drule_tac x=U in meta_spec) 
  apply(drule_tac x=V in meta_spec) 
  apply(drule_tac x=V in meta_spec) 
  apply(erule_tac x="i" in allE) 
  apply(erule_tac x="i" in allE)
  apply (force simp: T_Call);
  apply (force simp: T_Upd)
  apply (force simp: T_Active)
  apply (force simp: T_ActRef)
  apply (force simp: T_FutRef)
done

subsection{*substitution-related lemmas *}

lemma select_suc [simp] : "\<lbrakk>i < length list \<rbrakk> \<Longrightarrow>((a # list) ! (Suc i)) = list!i" 
apply(auto)
done

lemma subst_option_dom [simp] : "  dom (\<lambda>l. subst_option i s t (b l)) = (dom b)"
  apply(auto)
apply(case_tac "x \<in>  dom b")
apply(auto)
done


lemma subst_option : "\<forall> l \<in> dom b .  ( (the (b l))[s,t/x]) = the (subst_option x s t ( b l))"
  apply(auto)
done

lemma subst_lemma:
    " c,e \<turnstile> t : T  \<Longrightarrow> (\<And> e' i  U V u v. i \<le> length e' \<and>  c,e' \<turnstile> u : U \<and>  c,e' \<turnstile> v : V 
             \<Longrightarrow> e = e'\<langle>i:(U,V)\<rangle> \<Longrightarrow> c,e' \<turnstile> t[u,v/i] : T)"

apply(induct set: typing)

(* Self *)
 apply (rule_tac x = x and y = i in linorder_cases)
 apply simp;
 apply (rule_tac T'=T' in T_Self);
 apply(force)
 apply(simp)+;
 apply(case_tac x)
 apply(simp)
 apply(force simp: T_Self)
 
(* Param *)
 apply (rule_tac x = x and y = i in linorder_cases)
 apply simp;
 apply (rule_tac T'=T' in T_Param);
 apply(force)
 apply(simp)+;
 apply(case_tac x)
 apply(simp)
 apply(force simp: T_Param)

(* Obj *)
 apply(erule conjE)
 apply(simp)
 apply(rule typing.intros)
 apply(simp)
 apply(rule ballI)
 apply(simp add: subst_option)
 apply(erule_tac x=l in ballE)
 apply(erule conjE)+
 apply(erule_tac x=" e'\<langle>0:(B, fst B!l)\<rangle>" in allE)
 apply(erule_tac x="Suc i" in allE)
 apply(erule_tac x="U" in allE)
 apply(erule_tac x="V" in allE)
 apply(erule_tac x="(lift u 0)" in allE)
 apply(erule_tac x="(lift v 0)" in allE)
 apply(simp)+
 apply(erule impE)
 
 apply(drule_tac U=B and V=" fst B!l" in lift_type)
 apply(drule_tac U=B and V=" fst B!l" in lift_type)
 apply simp+;

 apply (rule_tac A=A in T_Call);
 apply(drule_tac x="e'" in meta_spec);
 apply(drule_tac x="i" in meta_spec);
 apply(drule_tac x="U" in meta_spec);
 apply(drule_tac x="V" in meta_spec);
 apply(drule_tac x=u in meta_spec);
 apply(drule_tac x=v in meta_spec);
 apply simp+;
 apply (rotate_tac 2);
 apply(drule_tac x="e'" in meta_spec);
 apply(drule_tac x="i" in meta_spec);
 apply(drule_tac x="U" in meta_spec);
 apply(drule_tac x="V" in meta_spec);
 apply(drule_tac x=u in meta_spec);
 apply(drule_tac x=v in meta_spec);
apply simp+;

 apply (rule T_Upd);
 apply(drule_tac x="e'" in meta_spec);
 apply(drule_tac x="i" in meta_spec);
 apply(drule_tac x="U" in meta_spec);
 apply(drule_tac x="V" in meta_spec);
 apply(drule_tac x=u in meta_spec);
 apply(drule_tac x=v in meta_spec);
 apply simp+;
 apply (rotate_tac 2);
 apply(drule_tac x="e'\<langle>0:(A, fst A!l)\<rangle>" in meta_spec);
 apply(drule_tac x="Suc i" in meta_spec);
 apply(drule_tac x="U" in meta_spec);
 apply(drule_tac x=V in meta_spec);
 apply(drule_tac x="lift u 0" in meta_spec);
 apply(drule_tac x="lift v 0" in meta_spec);
 apply simp;
 apply(simp add: lift_type);

 apply simp;
 apply (rule T_Active);
 apply(drule_tac x="e'" in meta_spec);
 apply(drule_tac x="i" in meta_spec);
 apply(drule_tac x="U" in meta_spec);
 apply(drule_tac x="V" in meta_spec);
 apply(drule_tac x=u in meta_spec);
 apply(drule_tac x=v in meta_spec);
 apply simp+;

 apply (simp add: T_ActRef T_FutRef)+; 
done;

lemma substs_lemma:
  "\<And>Ts. i \<le> length e \<and>  c,e \<turnstile> u : T\<and>  c,e \<turnstile> v : T' \<Longrightarrow> c,e\<langle>i:(T,T')\<rangle> ||- ts : Ts \<Longrightarrow>
     c,e ||- (map (\<lambda>t. t[u,v/i]) ts) : Ts"

  apply (induct ts)
  apply(simp)
  apply(simp)

  apply (case_tac Ts)
  apply simp
  apply simp
  apply atomize
  apply (case_tac Ts)
  apply simp
  apply simp
  apply (erule conjE)
  apply (rule subst_lemma);
  apply(auto)
done;  


subsection {* Subject reduction *}

lemma type_dom [simp] : "\<lbrakk>conf, env \<turnstile> (Obj a A) : A\<rbrakk> \<Longrightarrow> dom a = do A"
apply(erule typing_elims)
apply(auto)
done

lemma type_dom_2 : "\<lbrakk>conf, env \<turnstile> (Obj a A') : A\<rbrakk> \<Longrightarrow> dom a = do A"
apply(erule typing_elims)
apply(auto)
done

lemma select_preserve_type [simp] : 
" \<And> conf l1 l2 f env t T n A B A' B'. \<lbrakk>conf,env \<turnstile> Obj f T: Object t; the (t l2) = (A, B);
    conf,env\<langle>0:(Object t,A)\<rangle> \<turnstile> n : B;   l1 \<in> dom  t; l2 \<in>  dom  t; the (t l1) = (A',B') \<rbrakk>
       \<Longrightarrow> conf, env\<langle>0:(Object t,A')\<rangle> \<turnstile> (the  ((f(l2 |-> n)) l1)) : B'"

apply(case_tac "l1=l2")
apply(simp)
apply(subgoal_tac " the ((f(l2 \<mapsto> n)) l1) =  the (f l1)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac "l1 \<in> dom f")
apply(erule typing_elims)
apply(simp_all)
apply (erule_tac x=l1 in ballE);
apply (rotate_tac -2);
apply (drule sym);
apply simp+;
apply (erule typing_elims);
apply simp;
done;

lemma subject_reduction: "c,e \<turnstile> t : T \<Longrightarrow> (\<And>t'. t \<rightarrow>\<^sub>\<beta> t' \<Longrightarrow> c,e \<turnstile> t' : T)"
  apply (induct set: typing)
  apply (ind_cases "Var x \<rightarrow>\<^sub>\<beta> t'" for x t')
  apply (ind_cases "Var x \<rightarrow>\<^sub>\<beta> t'" for x t')
  apply (ind_cases "Obj b B \<rightarrow>\<^sub>\<beta> t'" for b B t')
  apply(subgoal_tac " the ((f(l \<mapsto> s)) l) = s")
  apply(subgoal_tac " \<forall>l\<in>do B.
           Tconf,env\<langle>0:(B, (fst B!l))\<rangle> \<turnstile> the (b l) : (snd B!l) \<and>
            (\<forall>x. the (b l) \<rightarrow>\<^sub>\<beta> x \<longrightarrow> Tconf,env\<langle>0:(B, (fst B!l))\<rangle> \<turnstile> x : (snd B!l))")
  apply(erule_tac x=l in ballE)
  prefer 2
  apply (simp,rule typing.intros);
  apply(force) 
  apply (rule ballI);apply (erule_tac x=la in ballE);prefer 2;apply simp+;
  apply (rule conjI);
  apply force;
  apply (rule impI);
  apply force;
  apply(erule conjE)
  apply(rotate_tac 4)
  apply (rotate_tac -3);
  apply(erule ssubst)

  apply(rule typing.intros)
  apply(subgoal_tac "dom (f(l \<mapsto> s)) = dom f")
  apply(simp)
  apply(rule dom_insert)
  apply(assumption)
  apply(rule ballI)
  apply(subgoal_tac "   \<forall>la\<in>do B.
           Tconf,env\<langle>0:(B, (fst B!la))\<rangle> \<turnstile> the ((f(l \<mapsto> s)) la) : (snd B!la) \<and>
           (\<forall>x. the ((f(l \<mapsto> s)) la) \<rightarrow>\<^sub>\<beta> x \<longrightarrow> Tconf,env\<langle>0:(B, (fst B!la))\<rangle> \<turnstile> x : (snd B!la))")
  apply(erule_tac x=la in ballE)
  apply(erule conjE)

  apply(case_tac "l = la")
  apply(subgoal_tac "the ((f(l \<mapsto> t)) la)= t")
  apply(simp)
  apply(rotate_tac -1)
  apply(erule ssubst)
  apply(rule insert_select)
  apply(erule_tac x=t in  allE)
  apply(erule impE)
  apply(simp)
  apply(subgoal_tac " \<forall>l\<in>do B.
           Tconf,env\<langle>0:(B, (fst B!l))\<rangle> \<turnstile> the (b l) : (snd B!l) \<and>
           (\<forall>x. the (b l) \<rightarrow>\<^sub>\<beta> x \<longrightarrow> Tconf,env\<langle>0:(B, (fst B!l))\<rangle> \<turnstile> x : (snd B!l))")
  apply(erule_tac x=la in ballE)
  apply(erule conjE)


  apply(subgoal_tac "the ((f(l \<mapsto> t)) la) = the (f la)")
  apply(subgoal_tac "the ((f(l \<mapsto> s)) la) = the (f la)")
  apply(simp)
  apply(rule the_insert_select)
  apply(subgoal_tac "la \<in> dom (f(l \<mapsto> s))")
  apply(subgoal_tac " dom (f(l \<mapsto> s)) = dom f")
  apply(simp)
  apply(rule dom_insert)
  apply(assumption)
  apply(simp)
  apply(assumption)
  apply(rule the_insert_select)
  apply(subgoal_tac "la \<in> dom (f(l \<mapsto> s))")
  apply(subgoal_tac " dom (f(l \<mapsto> s)) = dom f")
  apply(simp)
  apply(rule dom_insert)
  apply(assumption)
  apply(simp)
  apply(assumption)
  apply(erule_tac x=la in ballE)
  apply(simp)
  apply(simp)
  apply(rule ballI)

  apply(erule_tac x=lb in ballE)
  apply(rule conjI)
  apply(force)
  apply(simp)
  apply(simp)
  apply(erule_tac x=la in ballE)
  apply(erule conjE)
  apply(simp)+
    (* Obj done  - 5 *)

    (* now call *)
  apply (ind_cases "Call a l b \<rightarrow>\<^sub>\<beta> t' " for a l b t')
  apply(simp)
  apply(erule typing.cases)
  apply(simp)+
  apply (erule conjE)+;
  apply(rule_tac e= "enva\<langle>0:(Ba, (fst Ba!l))\<rangle>" and U="Ba" and V="(fst Ba!l)"  in  subst_lemma)
  apply simp+;
  apply (erule_tac x=l in ballE);
  apply simp+;
  apply (rule typing.intros);
  apply simp+;
  apply (rule_tac A = A in T_Call);
  apply simp+;
  apply (drule_tac x=t in meta_spec)+
  apply (rule_tac A = A in T_Call);
  apply simp+;

    (* now Upd - 4 *)
  apply (ind_cases "Upd a l n \<rightarrow>\<^sub>\<beta> t' " for a l n t')
  apply(erule typing.cases)
  apply(simp)+
  apply(erule conjE)
  apply(rotate_tac -1, drule sym)
  apply simp
  
  apply(rule typing.intros)
  apply(rotate_tac -4)
  apply(subgoal_tac  "l \<in> dom b")

  apply(erule subst)
  apply(rule dom_insert)
  apply(rotate_tac 1, drule sym)
  apply(simp)+

  apply (rule typing.intros,simp+)+;


    (*Active*)
  apply (ind_cases "Active a \<rightarrow>\<^sub>\<beta> t' " for a t')
  apply (erule ssubst)
  apply (rule T_Active);
  apply simp+;
  
  apply (ind_cases "Actref B \<rightarrow>\<^sub>\<beta> t' " for Actref B t',simp+);
  apply (ind_cases "Futref B \<rightarrow>\<^sub>\<beta> t' " for Futref B t',simp+);
done

theorem subject_reduction': "t \<rightarrow>\<^sub>\<beta>\<^sup>* t' \<Longrightarrow> c,e \<turnstile> t : T \<Longrightarrow> c,e \<turnstile> t' : T"
  by (induct set: rtranclp) (iprover intro: subject_reduction)+


subsection{* corolaries of typing: toward progress*}

lemma type_members_equal : "\<lbrakk>do A = do B; \<forall>i. A!!i = B!!i\<rbrakk> \<Longrightarrow> A = B"
apply(cases A)
apply(cases B)
apply(hypsubst)
apply(simp)
apply(rule ext)
apply(erule_tac x=x in allE)
apply(force)
done

lemma type_members_equal_dom : "\<lbrakk>do A = do B; \<forall> l \<in> do A. A!l = B!l\<rbrakk> \<Longrightarrow> A = B"
apply(cases A)
apply(cases B)
apply(hypsubst)
apply(simp)
apply(rule ext)
apply (case_tac "x : dom funa");
apply(erule_tac x=x in ballE)
apply(force)
apply (erule notE);
apply assumption;
apply (subgoal_tac "fun x = None")
apply (subgoal_tac "funa x = None")
apply auto;
done;

lemma not_var : "  \<lbrakk>c,[] \<turnstile> a : A \<rbrakk> \<Longrightarrow> \<forall> x.  a \<noteq> Var x"
apply (rule allI);
apply (case_tac x);
apply(auto)
done


lemma Call_label_range: "\<lbrakk> c,[] \<turnstile> Call (Obj ob U) l p : A; (c,[] \<turnstile> Obj ob U: T) \<rbrakk> \<Longrightarrow> l \<in>  dom ob";
apply (erule typing.cases);
apply simp+;
apply (erule conjE)+;
apply (rotate_tac -4);
apply (drule sym);
apply simp+;
apply (drule type_dom_2)+;
apply auto
done

lemma  Call_subterm_type: "\<lbrakk> c,env \<turnstile> Call t l u: T0 \<rbrakk> \<Longrightarrow> (\<exists> T. c,env \<turnstile> t : T)";
apply (erule typing.cases);
apply simp+;
apply (rule_tac x="A" in exI);
apply simp+;
done;

lemma  Call_subterm_type_2: "\<lbrakk> c,env \<turnstile> Call t l u: T0 \<rbrakk> \<Longrightarrow> (\<exists> T. c,env \<turnstile> u : T)";
  apply (erule typing.cases);
  apply simp+;
  apply (rule_tac x="B" in exI);
  apply simp+;
done;

lemma Upd_label_range: "!!x .  \<lbrakk> c,[] \<turnstile> Upd (Obj ob T) l x : A; c,[] \<turnstile> Obj ob T: T \<rbrakk> \<Longrightarrow>l \<in>  dom ob";
  apply (erule typing_elims);
  apply auto;
done;

lemma  Upd_subterm_type: "\<lbrakk> c,[] \<turnstile> Upd t l x : T0 \<rbrakk> \<Longrightarrow> (\<exists> T. c,[] \<turnstile> t : T) ";
  apply (erule typing.cases);
  apply simp+;
  apply (rule_tac x=A in exI);
  apply simp+;
done;

lemma no_var :" \<lbrakk>\<exists> T c. c,[] \<turnstile> Var x :T \<rbrakk> \<Longrightarrow> False"
  apply (case_tac x);
  apply auto;
done;

lemma type_Upd : "c,[] \<turnstile> Upd d1 l d2: A \<Longrightarrow> (\<exists> A. c,[] \<turnstile> d1: A)";
  apply (erule typing.cases);
  apply simp+;
  apply (rule_tac x=Aa in exI);
  apply simp+;
done;
 
lemma dom_Call_Obj : "c,[] \<turnstile> Call (Obj ob T) l u : A \<Longrightarrow> l: dom ob";
  apply (erule typing.cases);
  apply simp+;
  apply clarify;
  apply (rotate_tac 3);
  apply (drule sym);
  apply auto;
done;

lemma dom_Upd_Obj: "c,[] \<turnstile> Upd (Obj ob T) l d2 : T \<Longrightarrow> l: dom ob"; 
  apply (erule typing_elims);
  apply simp;
done;

lemma do_empty_eq: "dom empty = do T \<Longrightarrow> T = Object empty"; 
  apply simp;
  apply (case_tac T);
  apply simp;
  apply (rule ext);
  apply auto;
done;

lemma Obj_empty_unique: "\<lbrakk> c , env \<turnstile> Obj empty T: T; c , env \<turnstile> Obj empty T': T' \<rbrakk> \<Longrightarrow> T = T'";
  apply clarify;
  apply (drule do_empty_eq);
  apply (drule do_empty_eq);
  apply simp;
done;

lemma Object_type1: "(c, env \<turnstile> Obj a T: T) \<Longrightarrow> (? f. c, env \<turnstile> Obj a T: Object f)";
  apply (case_tac T);
by (rule_tac x="fun" in exI, simp);

theorem type_unique: "\<lbrakk> c, env \<turnstile> a: T; c, env \<turnstile> a: T' \<rbrakk> \<Longrightarrow> T = T'";

  apply (insert typing.induct [of c env a T "\<lambda> c env a T. c, env \<turnstile> a: T \<longrightarrow>
          (\<forall> T'. c, env \<turnstile> a: T' \<longrightarrow> T = T') "]);

  apply atomize;
  apply (erule impE)+;
  apply assumption;
  apply (erule impE);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply force;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply force;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply (erule impE);
  apply assumption;
  apply (erule impE);
  apply assumption;
  apply (erule typing_elims);
  apply (erule typing_elims);
  apply (frule_tac x=Aa in spec);
  apply (frule_tac x=Ab in spec);
  apply (erule impE);
  apply assumption;
  apply (erule impE);
  apply assumption;
  apply (rotate_tac -2);apply (drule sym);apply (drule sym);
  apply (hypsubst);
  apply (subgoal_tac "(Bb, T'a) = (Ba, Ta)");
  apply (rotate_tac -1);apply (drule_tac f=snd in arg_cong);
  apply (rotate_tac -1);apply (drule snd_eqD);
  apply (rotate_tac -1,drule sym,rotate_tac -1,drule snd_eqD);
  apply assumption;
  apply (rotate_tac -2,drule sym, rotate_tac -1,erule ssubst);
  apply (rotate_tac -4,drule sym, rotate_tac -1,erule ssubst);
  apply (simp(no_asm_use));

  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply (drule_tac x = T'a in spec);
  apply simp;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply (drule_tac x = T'a in spec);
  apply force;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply clarify;
  apply simp;
  apply (erule impE)+;
  apply ((rule allI)+,(rule impI)+);
  apply ((rule allI)+,(rule impI)+);
  apply (rotate_tac -3);
  apply (erule typing_elims);
  apply simp+;
  apply (erule typing_elims);
  apply simp+;
done;


lemma subterm_typing_pre[rule_format]: 
    "\<forall> T env'. c, env' \<turnstile> (E::general_context)\<up>a : T \<longrightarrow> (\<exists> T' env. c,env \<turnstile> a:T')";

apply (induct_tac E);
apply simp;
(* *)
apply (rule allI)+; apply (rule impI);
apply simp;
apply (erule typing.cases);
apply simp;
defer;
apply simp+;
prefer 5;
apply simp;
apply (erule conjE)+;
  (* obj *)
apply (erule impE);
apply (drule_tac x = "FLlabel FmapLabel" in bspec);
apply force;
apply (subgoal_tac "the (b (FLlabel FmapLabel))  = general_context \<up> a");
apply (rule_tac x = "snd B!FLlabel FmapLabel" in exI);
apply (rule_tac x = "env<0:(B, fst B!FLlabel FmapLabel)>" in exI);
apply simp;
apply force;
apply assumption;
  (* Call *)
apply simp+;
apply (rule impI);
apply (erule exE)+;
apply (erule typing.cases);
apply simp+;
apply (erule impE);
apply (rule_tac x = A in exI);
apply (erule_tac x = env in exI);
apply simp+;
(* Upd *)
apply (rule impI);
apply (erule exE)+;
apply (erule typing.cases);
apply simp+;
apply (erule impE);
apply (rule_tac x = B in exI);
apply (rule_tac x = env in exI);
apply simp+;
(* Upd *)
apply (rule impI);
apply (erule exE)+;
apply (erule typing.cases);
apply simp+;
apply (erule impE);
apply (rule_tac x = "A" in exI);
apply (rule_tac x = "env" in exI);
apply simp+;
apply (rule impI);
apply (erule exE)+;
apply (erule typing.cases);
apply simp+;
apply force;
apply simp+;
(* Active *)
apply (rule impI);
apply (erule exE)+;
apply (erule typing.cases);
apply simp+;
apply (erule impE);
apply (rule_tac x = A in exI);
apply (erule_tac x = env in exI);
apply simp+;
done;

lemma subterm_typing: "c,[] \<turnstile> (E::general_context)\<up>a : T \<Longrightarrow> (\<exists> T' env. c,env \<turnstile> a:T')";
  apply (erule subterm_typing_pre);
done;

lemma noFV_Var: "noFV (Var n) \<Longrightarrow> False";
  apply (unfold noFV_def);
  apply (case_tac n);
  apply simp+;
done;

lemma noFV_Call: "noFV (Call d l a) \<Longrightarrow> noFV d";
  apply (unfold noFV_def);
  apply simp;
done;

lemma noFV_Call_2: "noFV (Call d l a) \<Longrightarrow> noFV a";
  apply (unfold noFV_def);
  apply simp;
done;

lemma noFV_Upd: "noFV (Upd d l d') \<Longrightarrow> noFV d";
  apply (unfold noFV_def);
  apply simp;
done;


lemma fold_implies':" \<lbrakk>finite B;   Finite_Set.foldSet op \<and> P True B x\<rbrakk> \<Longrightarrow>(\<forall> x \<in>  B. P x) \<longrightarrow>  x"
  apply(erule foldSet.induct)
  apply(simp)+
done

lemma fold_implies:" \<lbrakk>finite B; (\<forall> x \<in>  B. P x);   Finite_Set.foldSet op \<and> P True B x\<rbrakk> \<Longrightarrow> x"
  apply(frule fold_implies')
  apply(simp)+
done

lemma  fold_lem: "\<lbrakk>finite B; (\<forall> x \<in>  B. P x)\<rbrakk> \<Longrightarrow> fold op \<and>  (\<lambda>  x. P x) True B ";
  apply(unfold fold_def)
  apply(frule_tac f="op \<and>" in  finite_imp_foldSet)
  apply(subst the1_equality)
  apply(auto)
  apply(rule fold_implies,(simp+));
  apply(rule fold_implies,(simp+));
  apply(frule fold_implies,(simp+));
done

lemma fold_True: "fold op \<and> (\<lambda>x. True) True (B::Label set)"
apply (rule fold_lem)
apply (rule finite_subset);
apply (rule subsetI);
apply (rule UNIV_I);
apply (rule finite_Label);
apply (rule ballI);
apply (rule TrueI);
done;

constdefs
   prefix :: "('a list) \<Rightarrow> ('a list) \<Rightarrow> bool"   
   "prefix l m ==  l = take (length l) m"

lemma prefix_implies :  "\<lbrakk> prefix l m \<rbrakk> \<Longrightarrow> \<forall> x < length l . l!x = m!x" 
apply  (unfold prefix_def)
apply(rule allI)
apply(rule impI)
apply(induct l)
apply(simp)+
done

lemma take_nth' :"\<lbrakk> a < length xs \<rbrakk>  \<Longrightarrow> (xs  ! a) = ((xs@[x]) ! a)"
by (subst  nth_append) simp;

lemma prefix_empty: "prefix [] l";
by (unfold prefix_def) simp;


lemma prefix_consistent: "\<lbrakk> prefix l l'; n < length l \<rbrakk> \<Longrightarrow> l ! n = l' ! n";
  apply (drule prefix_implies);
by auto;

lemma shift_prepend [simp] : "list<0:a> = a # list" 
  apply (unfold shift_def);
  apply (case_tac list);
  apply simp+;
done;

lemma prefix_cong: "prefix l l' \<Longrightarrow> prefix (l<0:B>)(l'<0:B>)";
  apply  (unfold prefix_def);
  apply simp;
done;

lemma nth_prefix : " \<lbrakk> prefix l m; x < length l\<rbrakk> \<Longrightarrow> l!x = m!x" 
  apply(frule prefix_implies)
  apply auto;
done


lemma prefix_length_trans: "\<lbrakk> x < length l; prefix l l' \<rbrakk> \<Longrightarrow> x < length l'";
  apply (unfold prefix_def);
  apply(subgoal_tac "length l = min (length l) (length l')")
  apply(subgoal_tac "length l \<le> length l'")
  apply(arith)+
  apply(subgoal_tac "length  (take (length l) l') = 
                     min (length  (take (length l) l')) (length l')")
  apply(simp add: length_take)
  apply(subst length_take)
  apply(simp)
done

lemma FVtest_Obj: "\<lbrakk> FVtest (Obj (f :: Label -~> dB) T) n; l: dom f \<rbrakk> \<Longrightarrow> FVtest (the (f l)) (Suc n)";

  apply simp;
  apply (subgoal_tac "op \<and> ((\<lambda>x. optionFVtest (f x) (Suc n)) l)(fold op \<and> (\<lambda>x. optionFVtest (f x) (Suc n)) True (dom f - {l}))");
  prefer 2;
  apply (subgoal_tac "ab_semigroup_mult (op &)");
  apply (frule fold_rec_forw);
  defer;
  apply assumption+;
  apply (simp add: and_ab_semigroup_mult);
  apply (subgoal_tac " \<exists> z. f l = Some z");
  apply (auto simp: finite_dom_fmap)
done;


lemma FVtest_UpdR: "FVtest (Upd aa l n) lt \<Longrightarrow> FVtest n (Suc lt)";
by simp;

lemma vartype_inv: "\<lbrakk> c, env \<turnstile> a: T; FVtest a (length list);  prefix list env \<rbrakk> 
                    \<Longrightarrow> c, list \<turnstile> a: T";

apply (insert typing.induct [of c env a T "\<lambda> c env a T. c, env \<turnstile> a: T \<longrightarrow>
 (\<forall> list. FVtest a (length list) \<longrightarrow> prefix list env \<longrightarrow> c, list \<turnstile> a: T)"]);

apply atomize;
apply (erule impE)+;
apply assumption;
apply (erule impE);
apply ((rule allI)+,(rule impI)+);
apply simp;
apply ((rule allI),(rule impI)+);
apply (rule T_Self);
apply simp+;
apply (rotate_tac -1);
apply (drule_tac n = x in prefix_consistent);
apply simp;
apply force;
apply (erule impE)+;
apply ((rule allI)+,(rule impI)+);
apply simp;
apply ((rule allI)+,(rule impI)+);
apply (rule T_Param);
apply simp+;
apply (rotate_tac -1);
apply (drule_tac n = x in prefix_consistent);
apply simp;
apply force;
(* *)
apply (erule impE)+;
apply ((rule allI)+,(rule impI)+);
apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply simp;
(* back to Obj *)
apply (rule T_Obj);
apply assumption;
apply (rule ballI);
apply (drule bspec);
apply assumption;
apply (erule conjE)+;
apply (erule impE);
apply force;
apply (drule_tac x = "lista<0:(B, fst B!l)>" in spec);
apply (erule impE);
apply (subgoal_tac "length (lista <0:(B, fst B!l)>) = Suc (length lista)");
apply (rotate_tac -1); apply (erule ssubst);
apply (erule FVtest_Obj);
apply simp;
apply simp;
apply (erule impE);
apply (erule prefix_cong);
apply assumption;
(* Call *)
apply (rule T_Call);
apply (erule impE);
apply assumption;
apply (drule_tac x = lista in spec);
apply (erule impE);
apply simp;
apply simp;
apply (erule impE);
apply assumption+;
apply simp+;
(* Upd *)
apply (rule T_Upd);
apply simp+;
apply clarify;
apply (rotate_tac 7,drule_tac x="(A, fst A!l) # lista" in spec);
apply (erule impE);
apply simp+;
apply (unfold prefix_def);
apply simp+;
(* Active *)
apply force;
apply force;
apply force;
done;


lemma NoFV_typing: "\<lbrakk> c, env \<turnstile> a: T;  noFV a \<rbrakk> \<Longrightarrow> c, [] \<turnstile> a:T";
  apply (erule vartype_inv);
  apply (unfold noFV_def, simp);
by (rule prefix_empty);

lemma vartype: " c,list \<turnstile> a : T \<Longrightarrow> FVtest a (length list)"
  apply (insert typing.induct [of c list a T "\<lambda>Tconf e a T .Tconf,e \<turnstile> a : T \<longrightarrow> FVtest a (length e) "]); 
  apply atomize;
  apply (erule impE)+;
  apply assumption;
  apply (erule impE);
  apply ((rule allI)+,(rule impI)+);
  apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;

  defer; (* Obj later*)
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply simp;
  apply simp;
  apply (subgoal_tac "\<forall>l\<in>do B.(FVtest (the (b l)) (Suc (length env)))");
  apply (subgoal_tac "fold (op \<and>) (\<lambda>x. optionFVtest (b x) (Suc (length env))) True (do B) = 
                     fold (op \<and>) (\<lambda>x. True) True (do B)");
  apply simp;
  defer;
  apply simp;
  apply (rule ab_semigroup_mult.fold_cong);
  apply (simp add: and_ab_semigroup_mult);
  defer;
  apply (rotate_tac -2);
  apply (drule_tac x=x in bspec);apply assumption;
  apply simp;
  apply (subgoal_tac " \<exists> a. b x = Some a");
  apply (erule exE);
  apply simp+;
  apply (subgoal_tac "x \<in>dom b");
  apply force;apply simp;
  apply (rule ballI);
  apply (drule_tac x=l in bspec);apply assumption;
  apply force;
  apply (rule fold_lem);
  apply (case_tac B);
  apply (simp add: finite_dom_fmap)
  apply force
  apply (subgoal_tac "do B\<subseteq>(UNIV::(Label set))");
  apply (insert finite_Label);
  apply (rule finite_subset);
  apply auto;
done;

lemma typing_noFV: "c,[] \<turnstile> a : T \<Longrightarrow> noFV a "
  apply (unfold noFV_def);
  apply (drule  vartype);
  apply simp;
done;

lemma vartype_invR: "\<lbrakk> c, list \<turnstile> a: T; FVtest a (length list);  prefix list env \<rbrakk> 
                    \<Longrightarrow> c, env \<turnstile> a: T";
apply (insert typing.induct [of c list a T "\<lambda> c list a T. c, list \<turnstile> a: T \<longrightarrow>
 (\<forall> env. FVtest a (length list) \<longrightarrow> prefix list env \<longrightarrow> c, env \<turnstile> a: T)"]);
apply atomize;
apply (erule impE)+;
apply assumption;
apply (erule impE);
apply ((rule allI)+,(rule impI)+);
apply simp;
apply ((rule allI),(rule impI)+);
apply (rule T_Self);
apply (erule prefix_length_trans);
apply assumption;
apply (rotate_tac -1);
apply (drule prefix_consistent);
apply assumption;
apply simp;
apply (erule impE);
apply ((rule allI)+,(rule impI)+);
apply simp;
apply ((rule allI),(rule impI)+);
apply (rule T_Param);
apply (erule prefix_length_trans);
apply assumption;
apply (rotate_tac -1);
apply (drule prefix_consistent);
apply assumption;
apply simp;
(* *)
apply (erule impE)+;
apply ((rule allI)+,(rule impI)+);
apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply (erule impE)+;apply ((rule allI)+,(rule impI)+);apply ((rule allI),(rule impI)+);
defer;
apply simp;
(* back to Obj *)
apply (rule T_Obj);
apply assumption;
apply (rule ballI);
apply (drule bspec);
apply assumption;
apply (erule conjE)+;
apply (erule impE);
apply assumption;
apply (drule_tac x = "envaa<0:(B, fst B!l)>" in spec);
apply (erule impE);
apply (subgoal_tac "length (enva <0: (B, fst B!l)>) = Suc (length enva)");
apply (rotate_tac -1); apply (erule ssubst);
apply (erule FVtest_Obj);
apply simp;
apply simp;
apply (erule impE);
apply (erule prefix_cong);
apply assumption;
(* Call *)
apply (rule T_Call);
apply (erule impE);
apply assumption;
apply (drule_tac x = envaa in spec);
apply (erule impE);
apply simp+;
(* Upd *)
apply (rule T_Upd);
apply simp+;
apply (rotate_tac 7);
apply (drule_tac x = "(A, fst A!l) # envaa" in spec);
apply (erule impE);
apply (unfold prefix_def);
apply simp+;
apply (fold prefix_def);
(* Active *)
apply force;
apply force;
apply force;
done;

lemma noVar_env: "c, [] \<turnstile> a: T \<Longrightarrow> c, env \<turnstile> a: T";
apply (rule vartype_invR);
apply assumption;
apply (erule vartype);
by (rule prefix_empty);


lemma
NoFV_subterm_type: "\<lbrakk>c,[] \<turnstile> (E::general_context)\<up>a : T; noFV a \<rbrakk> \<Longrightarrow> \<exists> T'. c, [] \<turnstile> a:T'"
apply (drule subterm_typing);
apply (erule exE)+;
apply (drule NoFV_typing,simp+);
apply (rule_tac x=T' in exI,simp);
done

lemma type_unique': "c, env \<turnstile> Obj f T: T' \<Longrightarrow> T = T'";
apply clarify;
done;


lemma
subst_term_typing_pre[rule_format]: 
"\<forall> T T' env. (c,env \<turnstile> (E::general_context)\<up>a : T) \<longrightarrow> (noFV a) \<longrightarrow> (noFV b) \<longrightarrow>
  (c, env \<turnstile> a:T') \<longrightarrow> (c, env \<turnstile> b:T')  \<longrightarrow> c, env \<turnstile> (E::general_context)\<up>b : T"

apply (induct_tac E); 
apply (rule allI)+; apply (rule impI)+;
apply simp;
apply (drule type_unique);
apply assumption;
apply (erule ssubst);
apply assumption;
(* *)
apply (rule allI)+;
apply simp;
apply (rule impI)+;
apply (frule type_unique');
apply simp;
apply (rule T_Obj);
apply force;
apply (rule ballI);
apply (case_tac "l = FLlabel FmapLabel");
apply (rotate_tac -1);
apply simp;
apply (subgoal_tac 
  "\<forall> l \<in> do T. c, env<0:(T, fst T!l)> \<turnstile> the ((FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> a)) l): snd T!l");
apply (rotate_tac -1);
apply (drule_tac x = l in bspec);
apply force;
apply simp;
apply (drule_tac x = "snd T!l" in spec);
apply (drule_tac x = "T'" in spec);
apply (drule_tac x = "env<0:(T, fst T!l)>" in spec);
apply simp;
apply (erule impE)+;
apply (rotate_tac -5); 
apply (drule NoFV_typing);
apply assumption;
apply (erule noVar_env);
apply (erule impE);
apply (rotate_tac -4);
apply (drule NoFV_typing);
apply assumption;
apply (erule noVar_env);
apply assumption;
apply force;
(* l ~= FLLabel FmapLabel *)
apply (subgoal_tac 
  "the ((FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> b)) l) = the ((FLmap FmapLabel) l)");
prefer 2;
apply force;
apply (erule ssubst);
apply (subgoal_tac 
  "the ((FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> a)) l) = the ((FLmap FmapLabel) l)");
prefer 2;
apply force;
apply (rotate_tac -1);
apply (subgoal_tac "\<forall> l \<in> do T. (c, env<0:(T, fst T!l)> \<turnstile> 
  the ((FLmap FmapLabel(FLlabel FmapLabel \<mapsto> general_context \<up> a)) l) : snd T!l)");
prefer 2;
apply force;
apply (drule_tac x = l in bspec);
apply assumption;
apply simp;
(* Call *)
apply (rule allI)+;
apply simp;
apply (rule impI)+;
apply (erule typing.cases);
apply simp+;
defer;
apply simp+;
prefer 5;
apply (rule T_Call);
apply (drule_tac x = A in spec);
apply (drule_tac x = "T'" in spec);
apply (drule_tac x = "enva" in spec);
apply (erule impE);
apply assumption;
apply (erule impE);
apply assumption;
apply (erule impE);
apply assumption+;
apply (rule allI)+;
apply (rule impI)+;
apply (erule typing.cases);
apply simp+;
apply clarify;
apply (rule T_Call);
apply simp+;
(* Upd *)
apply (rule allI)+;
apply (rule impI)+;
apply (rule T_Upd);
apply (erule typing.cases);
apply simp+;
apply force;
apply force;
(* UpdR *)
apply (rule allI)+;
apply (rule impI)+;
apply simp+;
apply (rule T_Upd);
apply force;
apply force;
apply (drule_tac x = "snd T!Label" in spec);
apply (drule_tac x = "T'" in spec);
apply (drule_tac x = "env<0:(T, fst T!Label)>" in spec);
apply (erule impE);
apply force;
apply (erule impE);
apply (rotate_tac -2); 
apply (drule NoFV_typing);
apply assumption;
apply (erule noVar_env);
apply (erule impE);
apply (rotate_tac -1);
apply (drule NoFV_typing);
apply assumption;
apply (erule noVar_env);
apply assumption;
(* *)
apply (rule allI)+;
apply (rule impI)+;
apply simp+;
apply (drule spec)+;
apply (rule T_Active);
apply (erule typing.cases);
apply simp+;
apply (erule impE,assumption)+;
apply simp+;
done;


lemma
subst_term_typing: "\<lbrakk>c,[] \<turnstile> (E::general_context)\<up>a : T; c, [] \<turnstile> a:T';c, [] \<turnstile> b:T' \<rbrakk> 
                    \<Longrightarrow> c,[] \<turnstile> (E::general_context)\<up>b : T";
  apply (rule subst_term_typing_pre);
  apply assumption+;
  apply (erule typing_noFV)+;
by assumption+;



lemma 
ext_amap_typing: "(TConfig amap futmap, e \<turnstile> a : T \<Longrightarrow>G\<notin>dom amap 
                  \<Longrightarrow>TConfig (amap(G\<mapsto>T')) futmap, e \<turnstile> a : T )"
apply (insert typing.induct [of "TConfig amap futmap" e a T "\<lambda>Tconf e a T . \<forall> amapa futmapa.( Tconf, e \<turnstile> a : T\<and>Tconf=TConfig amapa futmapa\<and>G\<notin>dom amapa\<longrightarrow> TConfig  (amapa(G\<mapsto>T')) futmapa, e \<turnstile>a:T)"]);
apply simp+;
apply atomize;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+);
apply force;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply force;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Obj);
apply force;
apply (rule ballI);
apply (erule_tac x=l in ballE);
apply (erule conjE);
apply force;
apply force;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Call);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Upd);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Active);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_ActRef);
apply simp+;
apply force;
apply simp+;
apply (force simp: T_FutRef);
done;


lemma 
ext_futmap_typing: "(TConfig amap futmap, e \<turnstile> a : T \<Longrightarrow>fi\<notin>dom futmap \<Longrightarrow>
                                       TConfig amap (futmap(fi\<mapsto>T')), e \<turnstile> a : T )"
apply (insert typing.induct [of "TConfig amap futmap" e a T "\<lambda>Tconf e a T . \<forall> amapa futmapa.( Tconf, e \<turnstile> a : T\<and>Tconf=TConfig amapa futmapa\<and>fi\<notin>dom futmapa\<longrightarrow> TConfig amapa (futmapa(fi\<mapsto>T')), e \<turnstile>a:T)"]);
apply simp+;
apply atomize;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+);
apply force;
apply (rotate_tac 2);
apply (drule mp); 
apply ((rule allI)+,(rule impI)+)+;
apply force;
apply (rotate_tac 2);
apply (drule mp); 
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Obj);
apply force;
apply (rule ballI);
apply (erule_tac x=l in ballE);
apply (erule conjE);
apply force;
apply force;
apply (rotate_tac 2);
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Call);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Upd);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_Active);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+)+;
apply (rule T_ActRef);
apply simp+;
apply (drule mp);
apply ((rule allI)+,(rule impI)+);
apply ((rule allI)+,(rule impI)+);
apply (rule T_FutRef);
apply simp+;
apply (rule impI);
apply simp+;
done;



subsection{*Subject reduction for configurations*}

lemma Union_eq: "A=B \<Longrightarrow> \<Union>A=\<Union>B"
by auto;

lemma Insert_eq: "A=B \<Longrightarrow> insert a A = insert a B"
by auto;

lemma Csubject_reduction1: "\<turnstile> c : CT \<Longrightarrow>(\<And>c' . c \<rightarrow>\<^sub>\<zeta> c' \<Longrightarrow> \<exists> CT'. \<turnstile> c' : CT')"
  apply (induct set: Ctyping);
  apply (frule_tac P=" \<exists>CT'. \<turnstile>c' : CT'" in zeta.cases);
  prefer 7;apply simp;

(* LOCAL REDUCTION *)
  apply (rule_tac x="TConfig amap futmap" in exI);
  apply (rule Ctyping.intros);
  (* dom act*)
  apply (drule  actdom_and_red_exact);
  apply force;
  (*dom fut*)
  apply (rotate_tac 1);
  apply simp;
  apply (rule Union_eq);
  apply (rule Collect_cong);
  apply (rule iffI);
  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;  
  apply (rule_tac x="if A=Aa then m(fk \<mapsto> a') else ma" in exI);
  apply simp+;
  apply (rule conjI);
  apply (rule impI);
  apply simp+;
  apply (rule conjI);
  apply force;
  apply (rule_tac x=A in exI);  
  apply simp+;
  apply (rule impI);
  apply (rule_tac x=Aa in exI);
  apply force;

  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;
  apply (case_tac "Aa=A");  
  apply simp+;
  apply (rule_tac x="m" in exI);
  apply (rule conjI);
  apply force;
  apply (rule_tac x=A in exI); 
  apply (rule_tac x=a'' in exI); 
  apply simp+;
  apply (rule_tac x="ma" in exI);
  apply force;

  (*typing*)
  apply (rule ballI);
  apply (erule_tac x=Aa in ballE); 
  apply (case_tac "Aa=A");
  apply (simp add: ActiveObject_def)

  apply (rule ballI);
  apply (erule conjE)+;
  apply (erule_tac x=fi in ballE);
  apply (case_tac "fi=fk"); 
  apply (simp add: OneRequest_def RequestQueue_def)+;
  apply (erule subject_reduction) apply assumption;
  apply (simp add: OneRequest_def RequestQueue_def)+;
  apply force;
  apply (simp add: ActiveObject_def)
  apply (simp add: OneRequest_def RequestQueue_def)+;
  apply force;


(* ACTIVE OBJECT CREATION *)
 (*dom act*)
  apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> ((E\<circ>cActive cHole) \<up> a) : the (futmap fk)");
  apply (frule NoFV_subterm_type);
  apply assumption;
  apply (erule exE);
  apply (rule_tac x="TConfig (amap(G \<mapsto> T')) futmap" in exI);
  apply (rule Ctyping.intros);
  apply simp+;apply (rule Insert_eq); 
  apply (rule sym, rule insert_absorb);apply force;


 (*dom fut*)
  apply simp+; 
  apply (subgoal_tac "\<Union>{dom ma |ma.
            \<exists>Aa. (Aa = A \<longrightarrow>
                  (A = G \<longrightarrow> empty = ma) \<and> (A \<noteq> G \<longrightarrow> m(fk \<mapsto> E \<up> ActRef G) = ma)) \<and>
                 (Aa \<noteq> A \<longrightarrow>
                  (Aa = G \<longrightarrow> empty = ma) \<and> (Aa \<noteq> G \<longrightarrow> (\<exists>a. ca Aa = Some (ma, a))))} =
\<Union>{dom ma |ma.
            \<exists>Aa. ((Aa = A \<longrightarrow>  ( m(fk \<mapsto> E \<up> ActRef G) = ma)) \<and>
                 (Aa \<noteq> A \<longrightarrow>  (\<exists>a. ca Aa = Some (ma, a))))}");
  apply (simp);
  apply (rule Union_eq);
  apply (rule Collect_cong);
  apply (rule iffI);
  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;  
  apply (rule_tac x="if Aa=A then m(fk \<mapsto> E \<up> ActRef G) else  ma" in exI);
  apply simp+;
  apply (rule conjI, rule impI);
  apply simp+;
  apply (rule conjI);
  apply (rule sym, rule insert_absorb);apply force;

  apply (rule_tac x=A in exI);  
  apply simp+;
  apply (rule impI);
  apply (rule_tac x=Aa in exI);
  apply force;

  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;
  apply (case_tac "Aa=A");  
  apply simp+;
  apply (rule_tac x="m" in exI);
  apply (rule conjI);
  apply (rotate_tac -2,drule sym,simp);
  apply (rule insert_absorb);apply force;
  apply (rule_tac x=A in exI); 
  apply (rule_tac x=a'' in exI); 
  apply simp+;
  apply (rule_tac x="ma" in exI);
  apply simp+;
  apply (erule exE)+;
 apply (rule_tac x=Aa in exI); 
  apply (rule_tac x=aa in exI); 
  apply simp;

  (* the subgoal !*)
  apply (rule set_ext);
  apply simp+;
  apply (rule iffI);
  apply (erule exE, erule conjE)+;
  apply simp;
  apply (case_tac "Aa=G");
  apply (subgoal_tac "A\<noteq>G")
  apply simp+;apply force;
  apply force;
  apply (rule_tac x="xa" in exI); 
  apply (rule conjI);
  apply (rule_tac x="ma" in exI);
  apply simp+;
  apply (rule_tac x=Aa in exI); 
  apply (rule conjI, rule impI);
  apply force;
  apply force;
  apply simp;

  apply (erule exE);
  apply ((erule conjE),(erule exE)?)+;
  apply (case_tac "Aa=G");
  apply force;
  apply simp+;
  apply (rule_tac x=xa in exI); 
  apply simp;
  apply (rule_tac x=ma in exI,simp); 
  apply (rule_tac x=Aa in exI,simp); 
  apply force;

  (*typing*)
  apply (subgoal_tac "A\<noteq>G");
  apply (rule ballI);
  apply simp+;
  apply (rule conjI,rule impI,rule conjI);
  apply simp+;
  apply (rule ext_amap_typing);
  apply (simp add: ActiveObject_def)
  apply (subgoal_tac "G \<notin>dom  ca");
  apply (drule sym);apply simp;apply force;
  
  (*futures in G*)
  apply (rule ballI);
  apply simp+;
  apply (simp add: OneRequest_def RequestQueue_def)+;

  apply (case_tac "Aa=G"); 
  apply force;
  apply (erule_tac x=Aa in ballE);
  prefer 2;
  apply (erule disjE);apply force;
  apply (erule disjE);apply force;
  apply simp;
  apply (case_tac "Aa=A");
  (* in A*)
  apply (rule conjI,(erule conjE)+);
  apply simp+;
  apply (rule conjI);
  apply (rule ext_amap_typing);
  apply (simp add: ActiveObject_def)
  apply (subgoal_tac "G\<notin> dom ca")
  apply simp+;
  apply force; 
  apply (rule conjI);

(*once*)
   apply (erule_tac x=fk in ballE);
  apply simp+;
  apply (drule_tac G=G and T'=T' in ext_amap_typing);
  apply simp;apply force;
  apply (drule_tac G=G and T'=T' in ext_amap_typing);apply simp;apply force;
  apply (subgoal_tac "TConfig (amap(G \<mapsto> T')) futmap , [] \<turnstile> Active a : T'"); 
  apply (drule_tac c="TConfig (amap(G \<mapsto> T')) futmap" 
                   and T'=T' and b ="ActRef G"  in subst_term_typing);
  apply (unfold noFV_def,force,fold noFV_def);
  apply (rule T_ActRef);
  apply simp+;
  apply (rule T_Active);
  apply ( assumption);
  apply force;

  apply (rule ballI);
  apply (rule conjI)
  apply (rule impI);
  
  (*twice: again the same proof :-( *)
   apply (erule_tac x=fk in ballE);
  apply simp+;
  apply (drule_tac G=G and T'=T' in ext_amap_typing);
  apply simp;apply force;
  apply (drule_tac G=G and T'=T' in ext_amap_typing);apply simp;apply force;
  apply (subgoal_tac "TConfig (amap(G \<mapsto> T')) futmap , [] \<turnstile> Active a : T'"); 
  apply (drule_tac c="TConfig (amap(G \<mapsto> T')) futmap" 
                   and T'=T' and b ="ActRef G"  in subst_term_typing);
  apply (unfold noFV_def,force,fold noFV_def);
  apply (rule T_ActRef);
  apply simp+;
  apply (rule T_Active);
  apply ( assumption);
  apply force;

  (* back to other cases *)
  apply (rule impI);  
  apply (erule_tac x=fi in ballE);
  apply (rotate_tac -1);
  apply (drule_tac G=G and T'=T' in ext_amap_typing);
  apply simp;apply force;
  apply assumption;
  apply force;
  apply (rule impI)+;
  apply (rule conjI);
  apply (erule conjE)+;
  apply (rotate_tac -2);
  apply (drule_tac G=G and T'=T' in ext_amap_typing);apply simp;apply force;
  apply (simp add: ActiveObject_def)
  apply simp+;

  apply ((erule conjE)+);
  apply (rule conjI);
  apply (rotate_tac -2);
  apply (drule_tac G=G and T'=T' in ext_amap_typing);apply simp;apply force;
  apply (simp add: ActiveObject_def)

  apply (rule ballI);
  apply (erule_tac x=fi in ballE);
  apply (rotate_tac -1);
  apply (drule_tac G=G and T'=T' in ext_amap_typing);apply simp;apply force;
  apply (simp add: ActiveObject_def)
  apply simp+;

  (* A\<noteq>G*)
  apply force;

  apply (erule_tac x=A in ballE);apply (erule conjE);
    apply (erule_tac x=fk in ballE);
  apply (simp add: OneRequest_def RequestQueue_def);
  apply (force simp: RequestQueue_def);
  apply force;

  (*REMOTE METHOD CALL *)
 (*dom act*)
  apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> (E \<up> (Call (ActRef B) li p)) : the (futmap fk)");
  apply (frule NoFV_subterm_type);
  apply simp+;
  apply (unfold noFV_def,simp+);
  apply (frule NoFV_subterm_type);
  apply (unfold noFV_def,simp);
  apply (erule exE);
  apply (thin_tac "\<exists>T'. TConfig amap futmap , [] \<turnstile> Call (ActRef B) li p : T'");
  apply (rule_tac x="TConfig (amap) (futmap(fn \<mapsto> T'))" in exI);
  apply (rule Ctyping.intros);
  apply simp+;
  apply (rule sym);
  apply (subgoal_tac "insert A' (dom c) = dom c");
  apply simp+
  apply (rule insert_absorb); 
  apply force;
  apply (rule insert_absorb);apply force;

 (*dom fut*)
  apply simp+; 
 apply (rule set_ext);
  apply simp+;
  apply (rule iffI);
  (*first implication*)
  apply ((erule disjE)+,simp+);
  apply (rule_tac x="dom (mb(fn \<mapsto> Call a li p))" in exI);
  apply simp+;
  apply (rule_tac x="mb(fn \<mapsto> Call a li p)" in exI);
  apply simp+;
  apply (rule_tac x="B" in exI);
  apply simp+;
  apply (erule exE,erule conjE)+;
  apply simp+;
  apply (erule exE)+;
  apply (rule_tac 
   x="dom(if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)) else if A=B then (mb(fn \<mapsto> Call a li p)) else m)"
    in exI);
  apply simp+;
  apply (rule conjI, rule impI)+;
  apply simp+;
  apply (rule_tac 
   x="if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)) else if A=B then (mb(fn \<mapsto> Call a li p)) else m"
    in exI);
  apply simp+;
  apply (rule_tac x=A in exI);
  apply force;
  apply (rule impI,rule conjI)+;
  apply (rule_tac 
   x="if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)) else if A=B then (mb(fn \<mapsto> Call a li p)) else m"
    in exI,simp+);
  apply (rule_tac x=A in exI,simp+);
  apply (rule impI);
  apply (rule_tac 
   x="if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)) else if A=B then (mb(fn \<mapsto> Call a li p)) else m"
    in exI,simp+);
  apply (rule_tac x=A in exI,simp+);

  (*other implication ...*)
  apply ((erule exE),(erule conjE)+)+;
  apply simp+;
  apply (case_tac "A=B"); 
  apply simp+;
  apply (rotate_tac 7,drule not_sym);
  apply (rule meson_disj_comm)
  apply clarify;
 apply (rule_tac x="dom mb" in exI,simp+,rule conjI);
   apply (rule_tac x="mb" in exI,simp+);
  apply (rule_tac x=B in exI,simp+);
  apply force;
  apply (rule disjI2);
  apply (rule_tac x=xa in exI,simp+);
  apply (case_tac "A=A'",simp+);
  apply (rule_tac x=m' in exI);
  apply (rule conjI);
  apply (rotate_tac -2,drule sym,simp);
  apply (rule insert_absorb);
  apply force;
  apply (rule_tac x=A in exI,simp+);

  apply (rule_tac x=m in exI,simp+);
  apply (rule_tac x=A in exI,simp);

  (*typing*)
  apply (subgoal_tac "fk\<noteq>fn");
  prefer 2;
  apply (rotate_tac 3)
  apply (erule_tac x=A' in ballE);
  apply force;apply force;

  apply (subgoal_tac "fn\<notin>dom futmap");prefer 2;
 (*prove fn not in dom futmap *)
  apply simp+;  apply (rule allI)+;
  apply (rule meson_disj_comm);apply clarify; apply simp+;
  apply (rotate_tac 3)
  apply (erule_tac x=Aa in ballE);
  apply force; apply force;
  (*done*)

  apply (subgoal_tac "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> E \<up> FutRef fn : the (futmap fk)"); 
  prefer 2;
  apply (subgoal_tac "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> FutRef fn : T'");prefer 2;
  apply (rule T_FutRef);apply force;apply force;
  apply (drule_tac fi=fn and T' = T' in ext_futmap_typing,assumption);
  apply simp; 
  apply (rule_tac a = "Call (ActRef B) li p" and T' = T' in  subst_term_typing);
  apply assumption;
  apply (rule ext_futmap_typing,simp);apply simp;apply simp;
 
  apply (rule ballI);
  apply (rule conjI);
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply (erule disjE,force);apply assumption;
  apply (erule conjE);
  apply (rule ext_futmap_typing);
  apply (subgoal_tac "c.\<^sub>A A = (c'.\<^sub>A A)");
  apply simp;
  apply (simp add: ActiveObject_def);
  apply assumption;

  apply (rule ballI);
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply (erule disjE,force);apply assumption;
  apply (erule conjE)+;
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply (erule disjE,force);apply assumption;

  apply (case_tac "A=B");
  apply (case_tac "fi=fn");
  apply (simp add: OneRequest_def RequestQueue_def);

  (* typing the new request : *)
     apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> ActRef B : the (amap B)");
     apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> a: the (amap B)");
     apply (subgoal_tac  "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> (cCallL cHole li p) \<up> a : T'");
     apply simp;
     apply (rule_tac a="ActRef B" and T'="the (amap B)" in subst_term_typing);
     apply (rule ext_futmap_typing,simp+);

     apply (rule T_ActRef);apply force; apply simp+;apply force;
     apply (rule ext_futmap_typing,(simp add: ActiveObject_def)+);
     apply (rule T_ActRef);apply force;apply force;
     apply simp+;

  apply (erule_tac x=fi in ballE);
  apply (subgoal_tac "(c'.\<^sub>FA<fi>)=(c.\<^sub>FA<fi>)");
  apply (rule ext_futmap_typing);
  apply simp+;
  apply (force simp:  OneRequest_def);
  apply (force simp: RequestQueue_def);

  apply (erule_tac x=fi in ballE);
  apply (case_tac "A=A'");
  apply (case_tac "fi=fk");
  apply (force simp: OneRequest_def RequestQueue_def);
  apply (subgoal_tac "(c'.\<^sub>FA<fi>)=(c.\<^sub>FA<fi>)");
  apply (rule ext_futmap_typing);
  apply simp+;
  apply (force simp:  RequestQueue_def);
  apply assumption
  apply (force simp:  OneRequest_def);


  apply (subgoal_tac "(c'.\<^sub>FA<fi>)=(c.\<^sub>FA<fi>)");
  apply (rule ext_futmap_typing);
  apply simp+;
  apply (subgoal_tac "fi\<noteq>fn");
  apply simp;
  apply (force simp:  RequestQueue_def);
  apply assumption
  apply (force simp:  OneRequest_def);

  apply (subgoal_tac "dom (c'.\<^sub>R A) = dom (c.\<^sub>R A)")
  apply ( simp add:  RequestQueue_def);
  apply (simp add: RequestQueue_def);
  apply (rule impI);apply (rule insert_absorb);apply force;


  (* AND FINALLY THE FIRST SUBGOAL !!! *)  
  apply (drule_tac x=A' in bspec);
  apply force;
  apply simp+;apply (erule conjE)+;
  apply (drule_tac x=fk in bspec);
  apply (force simp: RequestQueue_def);
  apply (simp add: OneRequest_def);

   (* self request !!!*)
 (*dom act*)
  apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> (E \<up> (Call (ActRef A') li p)) : the (futmap fk)");
  apply (frule NoFV_subterm_type);
  apply (unfold noFV_def,simp+);
  apply (frule NoFV_subterm_type);
  apply (unfold noFV_def,simp);
  apply (erule exE);
  apply (thin_tac "\<exists>T'. TConfig amap futmap , [] \<turnstile> Call (ActRef A') li p : T'");
  apply (rule_tac x="TConfig (amap) (futmap(fn \<mapsto> T'))" in exI);
  apply (rule Ctyping.intros);
  apply simp+;
  apply (rule sym);
  apply (subgoal_tac "insert A' (dom c) = dom c");
  apply simp+
  apply (rule insert_absorb); 
  apply force;

 (*dom fut*)
  apply simp+; 
 apply (rule set_ext);
  apply simp+;
  apply (rule iffI);
  (*first implication*)
  apply ((erule disjE)+,simp+);
  apply (rule_tac x="dom (m'(fk \<mapsto> E \<up> FutRef fn, fn \<mapsto> Call a li p))" in exI);
  apply simp+;
  apply (rule_tac x="m'(fk \<mapsto> E \<up> FutRef fn, fn \<mapsto> Call a li p)" in exI);
  apply simp+;
  apply (rule_tac x="A'" in exI);
  apply simp+;
  apply (erule exE,erule conjE)+;
  apply simp+;
  apply (erule exE)+;
  apply (rule_tac 
   x="dom(if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)(fn \<mapsto> Call a li p)) else m)"
    in exI);
  apply simp+;
  apply (rule conjI,rule impI)+;
  apply (rule conjI);
  apply (rule_tac 
   x="(if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)(fn \<mapsto> Call a li p)) else m)"
    in exI);
  apply simp+;
  apply (rule_tac x=A' in exI,rule impI);
  apply force;
  apply simp+;
  apply (rule impI)+;
  apply (rule_tac 
   x="if A=A' then (m'(fk \<mapsto> E \<up> FutRef fn)(fn \<mapsto> Call a li p)) else m"
    in exI);
  apply simp+;
  apply (rule_tac x=A in exI,simp+);

  (*other implication ...*)
  apply ((erule exE),(erule conjE)+)+;
  apply simp+;
  apply (case_tac "A=A'"); 
  apply simp+;
  apply (rule meson_disj_comm)
  apply clarify;
  apply (rule_tac x="dom m'" in exI,simp+,rule conjI);
  apply (rule_tac x="m'" in exI,simp+);
  apply (rule_tac x=A' in exI,simp+);
  apply(case_tac "x=fk",simp+);
  apply force;apply force;
  apply (rule disjI2);
  apply (rule_tac x=xa in exI,simp+);
  apply (rule_tac x=m in exI);
  apply (rule conjI);
  apply force;
  apply (rule_tac x=A in exI,simp);

  (*typing*)
  apply (subgoal_tac "fk\<noteq>fn");
  prefer 2;
  apply (rotate_tac 3)
  apply (erule_tac x=A' in ballE);
  apply force;apply force;

  apply (subgoal_tac "fn\<notin>dom futmap");prefer 2;
 (*prove fn not in dom futmap *)
  apply simp+;  apply (rule allI)+;
  apply (rule meson_disj_comm);apply clarify; apply simp+;
  apply (rotate_tac 3)
  apply (erule_tac x=Aa in ballE);
  apply force; apply force;
  (*done*)

  apply (subgoal_tac "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> E \<up> FutRef fn : the (futmap fk)"); 
  prefer 2;
  apply (subgoal_tac "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> FutRef fn : T'");prefer 2;
  apply (rule T_FutRef);apply force;apply force;
  apply (drule_tac fi=fn and T' = T' in ext_futmap_typing,assumption);
  apply simp; 
  apply (rule_tac a = "Call (ActRef A') li p" and T' = T' in  subst_term_typing);
  apply assumption;
  apply (rule ext_futmap_typing,simp);apply simp;apply simp;
 
  apply (rule ballI);
  apply (rule conjI);
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply assumption;
  apply (erule conjE);
  apply (rule ext_futmap_typing);
  apply (subgoal_tac "c.\<^sub>A A = (c'.\<^sub>A A)");
  apply simp;
  apply (simp add: ActiveObject_def);
  apply assumption;

  apply (rule ballI);
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply assumption;
  apply (erule conjE)+;
  apply (drule_tac x=A in bspec);
  (* A\<in> dom ca : *)  apply simp; apply (rotate_tac -1); apply (erule disjE,force);
                 apply assumption;

  apply (case_tac "A=A'");
  apply (case_tac "fi=fn");
  apply (simp add: OneRequest_def RequestQueue_def);

  (* typing the new request : *)
     apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> ActRef A' : the (amap A')");
     apply (subgoal_tac "TConfig amap futmap , [] \<turnstile> a: the (amap A')");
     apply (subgoal_tac  "TConfig amap (futmap(fn \<mapsto> T')) , [] \<turnstile> (cCallL cHole li p) \<up> a : T'");
     apply simp;
     apply (rule_tac a="ActRef A'" and T'="the (amap A')" in subst_term_typing);
     apply (rule ext_futmap_typing,simp+);

     apply (rule T_ActRef);apply force; apply simp+;apply force;
     apply (rule ext_futmap_typing,(simp add: ActiveObject_def)+);
     apply (rule T_ActRef);apply force;apply force;
     apply simp+;

  apply (erule_tac x=fi in ballE);
  apply (case_tac "fi=fk");
  apply (force simp: OneRequest_def RequestQueue_def);
  apply (subgoal_tac "(c'.\<^sub>FA<fi>)=(c.\<^sub>FA<fi>)");
  apply (rule ext_futmap_typing);
  apply simp+;
  apply (force simp:  OneRequest_def);
  apply (simp add: RequestQueue_def);
  apply force;

  apply (erule_tac x=fi in ballE);
  apply (subgoal_tac "(c'.\<^sub>FA<fi>)=(c.\<^sub>FA<fi>)");
  apply (rule ext_futmap_typing);
  apply simp+;
  apply (force simp:  RequestQueue_def);
  apply assumption
  apply (force simp:  OneRequest_def);

  apply (subgoal_tac "dom (c'.\<^sub>R A) = dom (c.\<^sub>R A)")
  apply ( simp add:  RequestQueue_def);
  apply (simp add: RequestQueue_def);


  (* AND FINALLY THE FIRST SUBGOAL !!! *)  
  apply (drule_tac x=A' in bspec);
  apply force;
  apply simp+;apply (erule conjE)+;
  apply (drule_tac x=fk in bspec);
  apply (force simp: RequestQueue_def);
  apply (simp add: OneRequest_def);

  (* AO UPDATE *)  
  apply (rule_tac x="TConfig (amap(G \<mapsto> (the (amap B)))) futmap" in exI);
  apply (rule Ctyping.intros);
 (*dom act*)
  apply simp+;
  apply (rule Insert_eq);
  apply (rule sym,rule insert_absorb);
  apply force;
 (*dom fut*)
  apply simp+; 
  apply (subgoal_tac "\<Union>{dom m |m.
            \<exists>A. (A = A' \<longrightarrow> (A' = G \<longrightarrow> empty = m) \<and> (A' \<noteq> G \<longrightarrow> m'(fk \<mapsto> E \<up> ActRef G) = m)) \<and>
                (A \<noteq> A' \<longrightarrow> (A = G \<longrightarrow> empty = m) \<and> (A \<noteq> G \<longrightarrow> (\<exists>a. c A = Some (m, a))))} =
\<Union>{dom m |m.
            \<exists>A. (A = A' \<longrightarrow> (m'(fk \<mapsto> E \<up> ActRef G) = m)) \<and>
                (A \<noteq> A' \<longrightarrow> (\<exists>a. c A = Some (m, a)))}");
  apply (simp);
  apply (rule Union_eq);
  apply (rule Collect_cong);
  apply (rule iffI);
  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;  
  apply (rule_tac x="if A=A' then m'(fk \<mapsto> E \<up> ActRef G) else  m" in exI);
  apply simp+;
  apply (rule conjI, rule impI);
  apply simp+;
  apply (rule conjI);
  apply (rule sym, rule insert_absorb);apply force;

  apply (rule_tac x=A in exI);  
  apply simp+;
  apply (rule impI);
  apply (rule_tac x=A in exI);
  apply force;

  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;
  apply (case_tac "A=A'");  
  apply simp+;
  apply (rule_tac x="m'" in exI);
  apply (rule conjI);
  apply (rotate_tac -2,drule sym,simp);
  apply (rule insert_absorb);
  apply force;
  apply (rule_tac x=A in exI); 
  apply (rule_tac x=a' in exI); 
  apply simp+;
  apply (rule_tac x="m" in exI);
  apply simp+;
  apply (erule exE)+;
  apply (rule_tac x=A in exI); 
  apply (rule_tac x=aa in exI); 
  apply simp;

  (* the subgoal !*)
  apply (rule set_ext);
  apply simp+;
  apply (rule iffI);
  apply (erule exE,erule conjE)+;
  apply (case_tac "A=G");
  apply (subgoal_tac "A'\<noteq>G")
  apply simp+;apply force;
  apply force;
  apply (rule_tac x="xa" in exI); 
  apply (rule conjI);
  apply (rule_tac x="m" in exI);
  apply simp+;
  apply (case_tac "A'=A");
  apply force;
  apply (rule_tac x="A" in exI);
  apply simp+;
  apply (erule exE)+;
  apply (rule_tac x=xa in exI); 
  apply simp;
  apply (erule conjE,erule exE)+;
  apply simp+;
  apply (case_tac "A=A'");
  apply simp+;
  apply (rule_tac x=m in exI,simp); 
  apply (rule_tac x=A in exI,simp); 
  apply force;
  apply simp+;
  apply (erule exE)+;
  apply (rule_tac x=m in exI,simp); 
  apply (rule_tac x=A in exI,simp); 
  apply force;

  (*typing*)
  apply (subgoal_tac "A'\<noteq>G");
  apply (subgoal_tac "B\<noteq>G");
  apply (subgoal_tac "li\<in>do (the (amap B))")
  (* this subgoal is useful  twice *)
  apply (subgoal_tac "TConfig (amap(G\<mapsto> the (amap B))) futmap , [] \<turnstile> Upd a li b: the (amap B)");
  apply (rule ballI);
  apply (case_tac "A=G");prefer 2;
  apply (frule_tac x=A in bspec);
  (* A\<in> dom c : *)apply simp; apply (erule disjE,force,force);
  apply simp; 
  apply (erule conjE,rule conjI);
  apply (subgoal_tac "c'.\<^sub>A A = (c .\<^sub>A A)");
  apply (rule ext_amap_typing) 
  apply simp+;apply force;
  apply (case_tac "A=A'",force simp: ActiveObject_def,force simp: ActiveObject_def);
  apply (rule ballI);
  apply (case_tac "A=A' \<and> fi = fk")
  apply (simp add: OneRequest_def)

  (* Main case: typing the ref to the new activity *)
	 apply (subgoal_tac "TConfig (amap(G\<mapsto> the (amap B))) futmap , [] \<turnstile> Upd (ActRef B) li b: the (amap B)");
	 apply (subgoal_tac "TConfig (amap(G\<mapsto> the (amap B))) futmap , [] \<turnstile> ActRef G: the (amap B)");
	 apply (erule conjE,drule_tac x=fk in bspec);
	 apply (force simp: RequestQueue_def);
	 apply (rotate_tac -1,drule_tac G=G and T'="the (amap B)" in ext_amap_typing);
	 apply (simp,force);
	 apply (rule_tac T="the (futmap fk)" and T'="the (amap B)" in subst_term_typing);
	 apply (simp add: noFV_def)+;
	 apply (rule T_ActRef);
	 apply simp+;
	 apply (rule T_Upd);
	 apply (rule T_ActRef);
	 apply simp+;
	 apply (subgoal_tac "B\<in> dom amap");
	 apply force;apply simp;apply force;
	 apply (subgoal_tac "B\<in> dom amap");
	 apply force;apply simp;apply force;
	 apply assumption;
	 apply (erule typing.cases);apply simp+;

  (* other cases *)
     apply (subgoal_tac "c'.\<^sub>FA<fi> = (c.\<^sub>FA<fi>)");
     apply (drule_tac x=fi in bspec);
     apply (case_tac "A=A'",force simp: RequestQueue_def,force simp: RequestQueue_def);
     apply ( rule ext_amap_typing)
     apply (force simp: OneRequest_def);
     apply (simp,force);
     apply (force simp: OneRequest_def);

  (* in G *)
     apply (rule conjI);
     apply ( simp add: ActiveObject_def);
   (*futures in G*)
     apply (rule ballI);
     apply (force simp: RequestQueue_def);
   
  (* Upd a li b : the (amap B) *)
	 apply (rule T_Upd);
	 apply (drule_tac x=B in bspec);
	 apply force; apply (erule conjE);
	 apply (rule ext_amap_typing);apply (simp add: ActiveObject_def);
	 apply simp;apply force;
	 apply assumption;

	 apply (rule ext_amap_typing);
	 apply simp;
	 apply (drule_tac x=A' in bspec);
	 apply force; apply (erule conjE)+;
	 apply (drule_tac x=fk in bspec);
	 apply (force simp: RequestQueue_def);
	 apply (simp add: OneRequest_def);
	 apply (drule NoFV_subterm_type);
	 apply (unfold noFV_def,simp);
	 apply (erule exE,subgoal_tac "T'= the (amap B)");
	 apply (rotate_tac -2,erule_tac typing.cases);
	 apply simp+;
	 apply force;
	 apply (simp,force);
  
  apply (drule_tac x=A' in bspec);
  apply (simp,force);
  apply (simp add: OneRequest_def);apply (erule conjE)+;
  apply (drule_tac x=fk in bspec);
  apply (force simp: RequestQueue_def);apply (rotate_tac -1);apply simp;
  apply (drule NoFV_subterm_type);	 apply (unfold noFV_def,simp);
  apply force;
  apply force;
  apply force;

   
   (* FUTURE UPDATE !!! *)
  apply (rule_tac x="TConfig amap futmap" in exI);
  apply simp+;
  apply (rule Ctyping.intros);
  (* dom amap *)
  apply force; 
  (* dom futmap VERY SIMILAR TO LOCAL RULE *)
  apply simp+;
  apply (rule Union_eq);
  apply (rule Collect_cong);
  apply (rule iffI);
  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;  
  apply (rule_tac x="if A=Aa then m(fk \<mapsto> E \<up> ai) else m" in exI);
  apply simp+;
  apply (rule conjI);
  apply (rule impI);
  apply simp+;
  apply (rule conjI);
  apply force;
  apply (rule_tac x=A in exI);  
  apply simp+;
  apply (rule impI);
  apply (rule_tac x=Aa in exI);
  apply force;

  apply (erule exE);
  apply (erule conjE)+;
  apply (erule exE)+;
  apply (case_tac "Aa=A");  
  apply simp+;
  apply (rule_tac x="ma" in exI);
  apply (rule conjI);
  apply force;
  apply (rule_tac x=A in exI); 
  apply (rule_tac x=a in exI); 
  apply simp+;
  apply (rule_tac x="m" in exI);
  apply force;

  (* typing *)
  apply (rule ballI);
  apply (subgoal_tac  "TConfig amap futmap , [] \<turnstile> ai : the (futmap fi)");
  apply (drule_tac x=Aa in bspec);
  apply (case_tac "Aa=A",force,force);
  apply (case_tac "Aa=A");
  apply (rule conjI);
  apply (simp add: ActiveObject_def);
  apply (rule ballI);
  apply simp+;
  apply (erule conjE)+;
  apply (drule_tac x=fia in bspec);
  apply (simp add: RequestQueue_def,erule disjE,force,force);
  
  apply (case_tac "fia=fk");
  apply (simp add: OneRequest_def);
  apply (rule_tac a="FutRef fi" and T'= "the (futmap fi)" in subst_term_typing);
  apply (assumption);
  apply (rule T_FutRef);
  apply (subgoal_tac "fi \<in> dom futmap");
  apply force; 
  (*proves "fi \<in> dom futmap"*) 
     apply simp; apply (rule_tac x= "dom mb" in exI);
     apply (rule conjI);apply (rule_tac x= " mb" in exI);apply (rule conjI);
     apply simp+;
     apply (rule_tac x= B in exI);  apply force;apply force;
  (* done *)
  (*proves "fi \<in> dom futmap"*) 
     apply simp; apply (rule_tac x= "dom mb" in exI);
     apply (rule conjI);apply (rule_tac x= " mb" in exI);apply (rule conjI);
     apply simp+;
     apply (rule_tac x= B in exI);  apply force;apply force;
  (* done *)
  apply simp+;
  (* other cases fia in Aa*)
     apply (subgoal_tac "c'.\<^sub>FA<fia> = (c.\<^sub>FA<fia>)");
     apply (force simp: OneRequest_def);
     apply (simp add: OneRequest_def);

  (* other cases outside Aa*)
     apply (rule conjI);
     apply (simp add: ActiveObject_def);
     apply (rule ballI);
     apply (erule conjE, drule_tac x=fia in bspec);
     apply (simp add: RequestQueue_def);
     apply (subgoal_tac "c'.\<^sub>FAa<fia> = (c.\<^sub>FAa<fia>)");
     apply (force simp: OneRequest_def);
     apply (simp add: OneRequest_def);

  apply simp+;
  apply (drule_tac x=B in bspec);
  apply force;
  apply (erule conjE)+;
  apply (drule_tac x=fi in bspec);
  apply (force simp: RequestQueue_def);
  apply (simp add: OneRequest_def);
done;

lemma Csubject_reduction: " c \<rightarrow>\<^sub>\<zeta>\<^sup>* c' \<Longrightarrow> \<turnstile> c : CT \<Longrightarrow>\<exists> CT'. \<turnstile> c' : CT'"
by  (induct set: rtranclp) (iprover intro: Csubject_reduction1)+

subsection{*Typing ensures well-formedness*}

lemma Typing_implies_WF1a: "\<turnstile> c : CT \<Longrightarrow> wf_1a c"
apply (unfold wf_1a_def);
apply auto;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply force;
apply (simp add: ActiveObject_def);
apply (erule conjE)+;
apply (drule NoFV_subterm_type);
apply (unfold noFV_def,force);
apply (erule exE);
apply (erule typing.cases);
apply auto;
done;

lemma Typing_implies_WF1b: "\<turnstile> c : CT \<Longrightarrow> wf_1b c"
apply (unfold wf_1b_def);
apply auto;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply force;
apply simp+;
apply (erule conjE)+;
apply (drule_tac x=fk in bspec);
apply force;
apply simp+;
apply (drule NoFV_subterm_type);
apply (unfold noFV_def,force);
apply (erule exE);
apply (rotate_tac -1,erule typing.cases);
apply simp+;apply force;apply simp+;
done;

lemma Typing_implies_WF2a: "\<turnstile> c : CT \<Longrightarrow> wf_2a c"
apply (unfold wf_2a_def);
apply auto;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply force;
apply simp+;
apply (erule conjE)+;
apply (drule NoFV_subterm_type);
apply (unfold noFV_def,force);
apply (erule exE);
apply (erule typing.cases);
apply simp+;
apply (erule exE,(erule conjE)+)+;
apply (erule exE)+;
apply (rule_tac x=Aa in bexI);
apply (auto simp: RequestQueue_def);
done;


lemma Typing_implies_WF2b: "\<turnstile> c : CT \<Longrightarrow> wf_2b c"
apply (unfold wf_2b_def);
apply auto;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply force;
apply simp+;
apply (erule conjE)+;
apply (drule_tac x=fk in bspec);
apply force;
apply simp+;
apply (drule NoFV_subterm_type);
apply (unfold noFV_def,force);
apply (erule exE);
apply (rotate_tac -1,erule typing.cases);
apply simp+;
apply (erule exE,(erule conjE)+)+;
apply (erule exE)+;
apply (rule_tac x=Aa in bexI);
apply (force simp: RequestQueue_def);
apply (simp,force);
done;


theorem Typing_implies_WF: "\<turnstile> c : CT \<Longrightarrow> wf_conf c"
apply (unfold wf_conf_def);
apply (simp add: Typing_implies_WF1a Typing_implies_WF1b Typing_implies_WF2a Typing_implies_WF2b);
done;


 (* a value is an AO ref or an object *) 
inductive_set values :: "dB set"
where
    value_ActRef: "(ActRef s) \<in> values"
|   value_Obj: "(Obj f T) \<in> values"
abbreviation "isvalue" :: "dB \<Rightarrow> bool"
where
 "isvalue d \<equiv> d \<in> values"
    

subsection{*Progress*}

inductive_set remotes :: "dB set"
  where
    remote_Request: "((E::econtext)\<up> (Call (ActRef B) l p)) \<in> remotes"
|   remote_FutRef: "((E::econtext)\<up> (FutRef fi)) \<in> remotes"
|   remote_active:  "((E::econtext)\<up> (Active a)) \<in> remotes"
|   remote_update:  "((E::econtext)\<up> (Upd (ActRef B) l a)) \<in> remotes"

abbreviation "isremote" :: "dB \<Rightarrow> bool"
where
 "isremote d \<equiv> d \<in> remotes"

lemma remote_call: "Call (ActRef B) l p \<in> remotes"
apply (subgoal_tac "Call (ActRef B) l p = eHole\<up> (Call (ActRef B) l p)");
apply (erule ssubst);apply (rule remote_Request);
apply (simp);
done;

lemma remote_context: "a\<in>remotes \<Longrightarrow>(E::econtext)\<up>a \<in> remotes"
apply (subgoal_tac "\<forall> (E::econtext) (F::econtext) x. E\<up>F\<up>x = (E\<circ>F)\<up>x");
prefer 2;  apply (force simp: eCompose_and_Fill);
apply (erule remotes.cases);
apply (simp,rule remote_Request);
apply (simp,rule remote_FutRef);
apply (simp,rule remote_active);
apply (simp,rule remote_update);
done;

lemma remote_context_inv: "(E::econtext)\<up>a \<notin> remotes \<Longrightarrow> a\<notin>remotes"
by (auto simp: remote_context); 

lemma pre_progress : "(\<forall>A Tc . (Tc,[]\<turnstile> t : A) \<longrightarrow> \<not>(isvalue t\<or> isremote t) \<longrightarrow> ( \<exists> b . ( t \<rightarrow>\<^sub>\<beta> b)))
  \<and> (\<forall>A Tc. Tc,[]\<turnstile> Obj f A: A \<longrightarrow> \<not>(isvalue(Obj f A)\<or> isremote (Obj f A)) \<longrightarrow> (\<exists> b . ( Obj f A \<rightarrow>\<^sub>\<beta> b)))";

apply (rule db_induct2)
apply(simp)
apply(rule impI)
apply(drule no_var)
apply simp;

apply (rule allI)+;
apply (rule impI)+;
apply (frule type_unique');
apply (drule_tac x = T in spec);
apply (drule_tac x = Tc in spec);
apply simp;

prefer 4;
apply (rule allI)+;
apply (rule impI)+;
apply (rotate_tac 4);
apply (erule contrapos_np);
apply (rule disjI1);
apply (rule values.intros);

(* Call *)
apply (rule allI)+;apply (rule impI)+;
apply (frule Call_subterm_type);
apply (erule exE);
apply (drule_tac x = "T" in spec);
apply (drule_tac x = "Tc" in spec);
apply (erule impE);
apply assumption;
apply (drule Hilbert_Choice.meson_not_disjD, erule conjE);
apply simp+;
apply (case_tac " \<exists> c T. d1 = Obj c T");
apply(erule exE)+;
apply (rule_tac x = "(the(c l))[(Obj c Ta),d2/0]" in exI);
apply hypsubst;
apply (rule beta);
apply (erule dom_Call_Obj);
apply simp;
apply (subgoal_tac "\<not>(d1\<in>values)");
prefer 2;
apply (subgoal_tac "\<forall> B. d1\<noteq>ActRef B");
apply clarify;
apply (erule values.cases);apply (drule_tac x=s in spec);
apply simp;
apply (drule_tac x=f in spec);apply simp;
apply (rule allI);
apply clarify;
apply (simp add: remote_call);

apply (subgoal_tac "d1 \<notin> remotes");
apply simp+;
apply (erule exE);
apply (rule_tac x="Call x l d2" in exI);
apply (rule beta.intros); apply assumption;
apply clarify;
apply (subgoal_tac "(eCall eHole l d2)\<up>d1\<in>remotes") 
apply simp;
apply (rule remote_context,assumption);

(* Upd *)
apply (rule allI)+;apply (rule impI)+;
apply (frule type_Upd);
apply (erule exE);
apply (drule_tac x = Aa in spec);
apply (drule_tac x = Tc in spec);
apply (erule impE);
apply (assumption);
apply (drule Hilbert_Choice.meson_not_disjD, erule conjE);
apply simp+;
apply (case_tac  "isvalue d1");
apply simp+;
apply (subgoal_tac " \<exists> f T. d1 = Obj f T");
apply (erule exE)+;
apply simp+;
apply (subgoal_tac "l \<in>dom f");
apply (drule_tac  a=d2 in beta.upd);
apply (rule_tac x= "Obj (f(l \<mapsto> d2)) T" in exI);
apply assumption;
apply (erule typing.cases);apply simp+;
apply (subgoal_tac "Tconf, [] \<turnstile> Obj f T: Ab");
prefer 2;
apply simp;
apply (rotate_tac -1);
apply (frule_tac type_unique');
apply (rotate_tac -1);
apply (subgoal_tac "dom f = do T");
apply (rotate_tac -1);
apply (erule ssubst);
apply simp;
apply force;
apply simp+;
apply (erule values.cases);
apply simp+;
apply (subgoal_tac "eHole \<up>(Upd (ActRef s) l d2) \<in> remotes");
apply simp;
apply (rule remote_update);
apply simp+;
apply (subgoal_tac "(eUpdL eHole l d2)\<up>d1\<notin>remotes");
apply (drule_tac remote_context_inv);
apply simp+;apply (erule exE);
apply (rule_tac x="Upd x l d2" in exI);
apply (rule beta.updL);
apply assumption;
apply force;

(* obj empty *)
apply (rule allI)+;apply (rule impI)+;
apply (erule contrapos_np);
apply (rule disjI1);apply (rule value_Obj);
(* active*)
apply (rule allI)+;
apply (rule impI)+;
apply (rotate_tac -1);
apply (erule contrapos_np);
apply (rule disjI2);
apply (subgoal_tac "Active dB = eHole \<up>(Active dB)")
apply (rotate_tac -1);apply (erule ssubst);
apply (rule remote_active);
apply (force simp: eCompose_and_Fill);
(* Actref *)
apply (rule allI)+;
apply (rule impI)+;
apply (rotate_tac -1);
apply (erule contrapos_np);
apply (rule disjI1);
apply (rule value_ActRef);
(* Futref *)
apply (rule allI)+;
apply (rule impI)+;
apply (rotate_tac -1);
apply (erule contrapos_np);
apply (rule disjI2);
apply (subgoal_tac "FutRef FutureRef = eHole \<up>(FutRef FutureRef)");
apply (erule ssubst);
apply (rule remote_FutRef);
apply (simp)
done;


theorem sigma_progress : " \<lbrakk>tc,[]\<turnstile> t : A; \<not>(isvalue t);\<not>(isremote t)\<rbrakk> \<Longrightarrow> (\<exists> b . ( t \<rightarrow>\<^sub>\<beta> b))  "
apply(simp add: pre_progress)
done

lemma econtext_typing_noFV[rule_format]: " \<forall>c a T. c,[]\<turnstile>(E::econtext)\<up>a:T\<longrightarrow>noFV a"
  apply (induct_tac E);
  apply simp+;
  apply clarify;
  apply (rule typing_noFV);apply simp;
  apply clarify;
  apply (erule typing.cases);
  apply simp+;
  apply (erule conjE)+;
  apply (erule_tac x=Tconf in allE);
  apply (erule_tac x=a in allE);
  apply force;
  apply simp+;
  apply clarify;
  apply (erule_tac x=c in allE);
  apply (erule_tac x=a in allE);
  apply force;
  apply clarify;
  apply (erule typing.cases);
  apply simp+;
  apply (erule_tac x=Tconf in allE);
  apply (erule_tac x=a in allE);
  apply force;
  apply simp+;
done;

lemma CallParam_typing_noFV: " c,[]\<turnstile>(E::econtext)\<up>(Call a l p):T\<Longrightarrow>noFV p"
  apply (drule_tac E=E in econtext_typing_noFV);
  apply (rule noFV_Call_2);
  apply simp;
done;

lemma Activate_lem: " \<lbrakk> finite_conf c; A\<in> dom c;fi\<in> dom (c.\<^sub>RA);
             TConfig amap futmap,[]\<turnstile> c.\<^sub>FA<fi> : T;  c.\<^sub>FA<fi> = (E::econtext) \<up> Active b\<rbrakk>
           \<Longrightarrow> \<exists>c'. c \<rightarrow>\<^sub>\<zeta>  c' \<and> (c'.\<^sub>FA<fi>)\<noteq> (c.\<^sub>FA<fi>)" 
apply (subgoal_tac "\<exists> (G::ActivityRef). G\<notin>dom c");
prefer 2; apply (rule fresh_actref_finite); apply (erule finite_confE1);
apply (erule exE)+;
apply (rule_tac x="c (A\<mapsto> ((c.\<^sub>RA) (fi \<mapsto> ((ectxt2gctxt E)\<up>ActRef G)),(c.\<^sub>AA)))(G\<mapsto>(empty,b))" in exI);
apply simp;
apply (rule conjI);
apply (rule_tac zeta.active);
apply force;
apply (erule ActivityDef);
apply (force simp: RequestQueue_def OneRequest_def ectxt2gctxt_and_fill);
apply (rule_tac c="TConfig amap futmap" and E = "E\<circ>(eActive eHole)" and T=T in econtext_typing_noFV);
apply (simp add: eCompose_and_Fill)+;
apply (simp add: RequestQueue_def OneRequest_def);
apply (rule conjI);apply force;
apply (rule impI);
apply (simp add: ectxt2gctxt_and_fill);
apply clarify; 
apply (drule eq_econtext,simp);
done;



lemma cycle_c_00: "\<lbrakk> B : dom c; fia \<in> dom (c .\<^sub>R B); (c.\<^sub>FB<fia>) = (FutRef fia)\<rbrakk>
                   \<Longrightarrow> (fia,fia) : fcontainsf c";
apply (unfold fcontainsf_def);
apply simp;
apply (rule_tac x = B in bexI);
prefer 2;
apply assumption;
apply (erule conjI);
apply simp;
apply (rule_tac x = cHole in exI);
apply simp;
done;

lemma fcontainsf_subset_knows: "(f, f'): (fcontainsf c ) \<Longrightarrow> ((RefFut f, RefFut f'): knows c)";
apply (unfold fcontainsf_def knows_def, simp);
done;

lemma cycle_c_01: "\<lbrakk> B : dom c; fia \<in> dom (c .\<^sub>R B); (c.\<^sub>FB<fia>) = (FutRef fia)\<rbrakk>
                   \<Longrightarrow> (RefFut fia, RefFut fia) : knows c";
apply (rule fcontainsf_subset_knows);
apply (rule cycle_c_00);
apply assumption+;
done;

lemma cycle_c_0: "\<lbrakk> B : dom c; fia \<in> dom (c .\<^sub>R B); (c.\<^sub>FB<fia>) = (FutRef fia)\<rbrakk>
                   \<Longrightarrow> (RefFut fia, RefFut fia) : (knows c)^+";
apply (drule cycle_c_01, simp);
apply assumption+;
apply simp;
done;

lemma cycle_c: "\<lbrakk> B : dom c; fia \<in> dom (c .\<^sub>R B); (c.\<^sub>FB<fia>) = (FutRef fia)\<rbrakk>
                   \<Longrightarrow> ~(nocycle c)";
apply (unfold nocycle_def);
apply simp;
apply (drule_tac cycle_c_0);
apply assumption+;
apply (erule_tac x = "RefFut fia" in exI);
done;

lemma nocycle_fut_ref0: "\<lbrakk> nocycle c; B : dom c; fia \<in> dom (c .\<^sub>R B) \<rbrakk>
                           \<Longrightarrow> ~((c.\<^sub>FB<fia>) = (FutRef fia))";
apply (rule notI);
apply (drule cycle_c);
apply assumption+;
(* apply (insert all_nocycle); *)
apply (erule notE);
apply assumption;
done;

lemma nocycle_fut_ref: "\<lbrakk> nocycle c; B : dom c; fia \<in> dom (c .\<^sub>R B) \<rbrakk>
                        \<Longrightarrow> ~(E \<up> (c.\<^sub>FB<fia>) = (E :: econtext)\<up>(FutRef fia))";
apply (rule notI);
thm eq_context;
apply (drule eq_econtext);
apply (insert nocycle_fut_ref0);
apply simp;
done;


lemma remote_progress: " \<lbrakk> nocycle c; finite_conf c; \<turnstile> c : TConfig amap futmap; 
                           A\<in> dom c;fi\<in> dom (c.\<^sub>RA); (isremote (c.\<^sub>FA<fi>)) 
                         \<rbrakk>\<Longrightarrow> \<exists> c' . (( c \<rightarrow>\<^sub>\<zeta> c')\<and> ((c'.\<^sub>FA<fi>)\<noteq> (c.\<^sub>FA<fi>)))"
apply (subgoal_tac " TConfig amap futmap,[]\<turnstile> (c.\<^sub>FA<fi>) : the (futmap fi)");
apply (subgoal_tac "wf_conf c");
apply (erule remotes.cases);
apply (subgoal_tac "\<exists> (f::FutureRef). f\<notin>\<Union>{dom (fst (the (c A)))|A. A\<in>dom c}");
prefer 2; apply (rule fresh_futref_finite);
apply (erule finite_conf_unionE); 
apply (erule exE);
apply (case_tac "A=B");prefer 2; (*self request later *)

(* can do REQUEST *)
apply (rule_tac x=" c (A\<mapsto> ((c.\<^sub>RA) (fi \<mapsto> ((ectxt2gctxt E)\<up>(FutRef(f)))), c.\<^sub>AA))
                      (B \<mapsto> ((c.\<^sub>RB)(f \<mapsto> (Call (c.\<^sub>AB) l p)), (c.\<^sub>AB)))" in exI);
apply (rule conjI)
apply (rule zeta.request);
apply (rule ballI);
apply simp+;
apply (erule_tac x="dom (fst (the (c Aa)))" in allE);
apply (erule disjE);
apply (erule_tac x=Aa in allE);
apply (simp add: ectxt2gctxt_and_fill)+;
apply (rule ActivityDef)
apply force;
apply (simp add: ectxt2gctxt_and_fill RequestQueue_def OneRequest_def);
apply force;
apply (rule ActivityDef)
apply (unfold wf_conf_def wf_1b_def);
apply (erule conjE)+;
apply (rotate_tac -3,drule_tac x=A in bspec);
apply force;
apply (drule_tac x=fi in bspec);
apply assumption
apply (drule_tac x="(ectxt2gctxt E)\<circ>(cCallL cHole l p)" in spec);
apply (drule_tac x=B in spec);
apply (simp add: ectxt2gctxt_and_fill  OneRequest_def);
apply assumption
apply (rule_tac E=E and a="ActRef B" and l=l and T="the (futmap fi)" in CallParam_typing_noFV);
apply simp;

 (* c'.\<^sub>FA<fi>  \<noteq>  c.\<^sub>FA<fi>*)
  apply (simp add: OneRequest_def ectxt2gctxt_and_fill);
  apply (clarify, drule eq_econtext,force);

(*Self request*)
apply (rule_tac x="c (A\<mapsto> ((c.\<^sub>RA) (fi \<mapsto> ((ectxt2gctxt E)\<up>(FutRef(f))))(f \<mapsto> (Call (c.\<^sub>AA) l p)),
                                 (c.\<^sub>AA)))" in exI);
apply (rule conjI);
apply (rule zeta.request_self);
apply (rule ballI);
apply simp+;
apply (erule_tac x="dom (fst (the (c Aa)))" in allE);
apply (erule disjE);
apply (erule_tac x=Aa in allE);
apply simp+;
apply (rule ActivityDef)
apply force;
apply (simp add: ectxt2gctxt_and_fill RequestQueue_def OneRequest_def)+;
apply force;
apply (rule_tac E=E and a="ActRef B" and l=l and T="the (futmap fi)" in CallParam_typing_noFV);
apply simp;
 (* c'.\<^sub>FA<fi>  \<noteq>  c.\<^sub>FA<fi>*)
  apply (subgoal_tac "fi\<noteq>f");
  apply (simp add: OneRequest_def ectxt2gctxt_and_fill);
  apply (clarify, drule eq_econtext,force);
  apply (subgoal_tac "fi\<in> dom (c.\<^sub>RA)");
  apply (simp add: RequestQueue_def);
  apply (erule_tac x="dom (fst (the (c A)))" in allE);
  apply simp+;
  apply (erule disjE)
  apply (erule_tac x=A in allE);
  apply simp;
  apply (clarify,simp,force);
apply (fold wf_1b_def);

(* REPLY *)
apply (unfold wf_2b_def);
apply (erule conjE)+;
apply (drule_tac x=A in bspec);
apply force;
apply (fold wf_2b_def);
apply (drule_tac x=fi in bspec);
apply simp+;
apply (drule_tac x="ectxt2gctxt E" in spec);
apply (simp add: ectxt2gctxt_and_fill)+;
apply (drule_tac x=fia in spec);
apply simp;
apply (erule bexE);
apply (rule_tac x="c (A\<mapsto> (( c .\<^sub>RA)(fi \<mapsto> ((ectxt2gctxt E)\<up>((c.\<^sub>FB<fia>)))),( c .\<^sub>AA)))" in exI);
apply (rule conjI)
apply (rule_tac B=B in zeta.reply);
apply (rule ActivityDef);
apply assumption;
apply (force simp: ectxt2gctxt_and_fill RequestQueue_def OneRequest_def);
apply (rule ActivityDef);
apply assumption;
apply (force simp: RequestQueue_def OneRequest_def);

 (* c'.\<^sub>FA<fi>  \<noteq>  c.\<^sub>FA<fi>*)
  apply (simp add: OneRequest_def ectxt2gctxt_and_fill);
  apply (insert nocycle_fut_ref);
  apply (simp add: OneRequest_def);
(* done using above nocycle hypothesis.
  previous comment:
 come back here with no cycle of futures
 apply (clarify, drule eq_econtext,force);
 *)

(*ACTIVE direct*)
apply (frule_tac amap = amap and E=E and A=A and fi = fi  in Activate_lem);
apply simp+;

(* UPDATE *)
apply (subgoal_tac "\<exists> (G::ActivityRef). G\<notin>dom c");
prefer 2; apply (rule fresh_actref_finite);apply (erule finite_confE1);
apply (erule exE);
apply (rule_tac x="c (A\<mapsto> ((c .\<^sub>R A) (fi \<mapsto> ((ectxt2gctxt E)\<up>(ActRef G))),(c .\<^sub>A A)))
                                     (G \<mapsto> (empty, Upd (c .\<^sub>A B) l  a))" in exI);
apply (rule conjI);
apply (rule_tac B=B in zeta.update_AO);
apply force;
apply (erule ActivityDef);
apply (force simp: ectxt2gctxt_and_fill RequestQueue_def OneRequest_def);
apply (unfold wf_1b_def);
apply (erule conjE)+;
apply (drule_tac x=A in bspec);
apply force;
apply (fold wf_1b_def);
apply (drule_tac x=fi in bspec);
apply simp+;
apply (drule_tac x="(ectxt2gctxt E)\<circ>(cUpdL cHole l a)" in spec);
apply (simp add: ectxt2gctxt_and_fill)+;
apply (drule_tac x=B in spec);
apply (rule ActivityDef)
apply (simp add: ectxt2gctxt_and_fill RequestQueue_def OneRequest_def);
apply (rule_tac c="TConfig amap futmap" in  econtext_typing_noFV);
apply simp+;

 (* c'.\<^sub>FA<fi>  \<noteq>  c.\<^sub>FA<fi>*)
  apply (simp add: OneRequest_def ectxt2gctxt_and_fill);
  apply (subgoal_tac "A\<noteq>G");
  apply (simp,clarify,simp);
  apply (drule eq_econtext);
  apply force;
  apply (clarify,force);

(* subgoals *)
apply (fold wf_conf_def);
apply (rule Typing_implies_WF);
apply simp+;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply simp+;
done;

theorem progress_ASP : " \<lbrakk>nocycle c; finite_conf c; \<turnstile> c :TC; A\<in> dom c;fi\<in>dom (c.\<^sub>RA);\<not> (isvalue (c.\<^sub>FA<fi>))\<rbrakk>
                               \<Longrightarrow> \<exists> c' . (( c \<rightarrow>\<^sub>\<zeta> c')\<and> ((c'.\<^sub>FA<fi>)\<noteq> (c.\<^sub>FA<fi>) |
                                                        (c.\<^sub>FA<fi>) \<rightarrow>\<^sub>\<beta> (c.\<^sub>FA<fi>)))"
apply (case_tac "isremote (c.\<^sub>FA<fi>)");
apply (case_tac TC);
apply (drule remote_progress);
apply simp+;
apply force;
apply (erule Ctyping.cases);
apply (drule_tac x=A in bspec);
apply simp+;apply (erule conjE)+;
apply (drule_tac x=fi in bspec);
apply force;
apply (rotate_tac -1);
apply (drule sigma_progress);
apply simp+;
apply (erule exE);
apply (rule_tac x="ca (A\<mapsto> ((c.\<^sub>RA) (fi \<mapsto> b),(c.\<^sub>AA)))" in exI);
apply (rule conjI);
apply (rule zeta.zeta_local);
apply (simp)+;
apply (rule ActivityDef);apply assumption;
apply (force simp: RequestQueue_def OneRequest_def);
apply (simp add: RequestQueue_def OneRequest_def);
apply force;
done;

theorem progress_ASP2 : " \<lbrakk>nocycle c; finite_conf c;\<turnstile> c : TC; 
                           \<not>(\<exists> c'.(c \<rightarrow>\<^sub>\<zeta> c'));A\<in> dom c;fi\<in>dom (c.\<^sub>RA)\<rbrakk>
                                                \<Longrightarrow>c.\<^sub>FA<fi>\<in>values"
  apply (erule contrapos_np);
  apply (drule_tac A=A in progress_ASP);
  apply auto;
done;

theorem progress_ASP_reach[rule_format] : " \<lbrakk>c: reachable_configs;
                   \<turnstile> c :TC; A\<in> dom c;fi\<in>dom (c.\<^sub>RA);\<not> (isvalue (c.\<^sub>FA<fi>))\<rbrakk>
                               \<Longrightarrow> \<exists> c' . (( c \<rightarrow>\<^sub>\<zeta> c')\<and> ((c'.\<^sub>FA<fi>)\<noteq> (c.\<^sub>FA<fi>) |
                                                        (c.\<^sub>FA<fi>) \<rightarrow>\<^sub>\<beta>  (c.\<^sub>FA<fi>)))";
  apply (rule progress_ASP);
  apply (erule reachable_nocycle);
  apply (insert finite_reachable_configs);
  apply (erule_tac x = c in bspec);
  apply assumption+;
done;

lemma typing_empty_noRef: "TConfig empty empty, E \<turnstile> a : T \<Longrightarrow>noRef a "
  apply (insert typing.induct [of "TConfig empty empty" E a T "\<lambda>C L a T .C, L  \<turnstile> a : T \<longrightarrow> C=TConfig empty empty\<longrightarrow> noRef a"]); 
apply atomize;
apply (erule impE,simp)+;
apply clarify;
  apply (rule fold_lem);  apply (case_tac B);
  apply (simp add: finite_dom_fmap)
apply simp+;
apply clarify;
apply (drule_tac x=x in bspec,simp)+;
apply (case_tac "b x");
apply simp;apply simp;
apply (erule impE,simp)+;
apply clarify;
apply simp+;
apply clarify;
done;

lemma typed_init_config: "TConfig empty empty, [] \<turnstile> a : T\<Longrightarrow> \<turnstile>init_config a:TConfig [aref 0\<mapsto>Object empty] [fref 0\<mapsto>T] "
apply (unfold init_config_def)
  apply (rule Ctyping.intros);
apply simp+;
apply rule;
  apply (simp add: ActiveObject_def)
apply (rule T_Obj);
apply simp+;
apply rule;
apply rule;
apply clarify;
  apply (simp add: RequestQueue_def)
  apply (simp add: OneRequest_def);
apply (force simp: ext_amap_typing ext_futmap_typing);
apply clarify;
  apply (simp add: RequestQueue_def)
done;

theorem progress_ASP_init_conf : " \<lbrakk>init_config a \<rightarrow>\<^sub>\<zeta>\<^sup>* c;TConfig empty empty, [] \<turnstile> a : T;
                    A\<in> dom c;fi\<in>dom (c.\<^sub>RA)\<rbrakk>
                               \<Longrightarrow> (isvalue (c.\<^sub>FA<fi>)) \<or> (\<exists> c' . (( c \<rightarrow>\<^sub>\<zeta> c')\<and> ((c'.\<^sub>FA<fi>)\<noteq> (c.\<^sub>FA<fi>) |
                                                        (c.\<^sub>FA<fi>) \<rightarrow>\<^sub>\<beta>  (c.\<^sub>FA<fi>))))";
apply (case_tac "(isvalue (c.\<^sub>FA<fi>))");
apply simp;
apply (rule disjI2);
apply (subgoal_tac "\<exists> TC. \<turnstile>c:TC")
apply (erule exE);
apply (rule progress_ASP_reach);
apply (unfold reachable_configs_def init_config_def)
apply simp+;
apply (drule typing_empty_noRef)
apply force;
apply simp+;
apply (drule typed_init_config);
apply (drule_tac c'=c in Csubject_reduction);
apply (unfold init_config_def)
apply simp+;
done;
end
